#!/usr/bin/env python

#############################################################################
##
## This file is part of Taurus, a Tango User Interface Library
## 
## http://www.tango-controls.org/static/taurus/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
## 
## Taurus is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
## 
## Taurus is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
## 
## You should have received a copy of the GNU Lesser General Public License
## along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
###########################################################################

"""
TaurusGUI configuration file used to convert Taurus into VACCA

This file should be NOT modified.

.. toctree::
   :maxdepth: 4

h2. Usage
=========

This configuration file determines the default, permanent, pre-defined
contents of the GUI. While the user may add/remove more elements at run
time and those customizations will also be stored, this file defines what a
user will find when launching the GUI for the first time.

Don't modify this file, use default.py to do a simple setup of your
facility and paths and a CONFIG.py file (based on default.py format) to 
construct a VACCA gui for each subsystem.

h3. Contents
------------

This file mainly defines two kinds of objects:

 * Panel descriptors: the panels to be shown.
   To define a panel, instantiate a PanelDescription object (see documentation
   for the gblgui_utils module)
 * Panels will be TREE, JDRAW_FILE, DEVICE, ...
   
 * QSignals (managed by Taurus Shared Data Manager):
  * SelectedInstrument: goes from taurusgui and tree into panel/synoptic
  * SelectionMultiple: goes from panic/finder/... to synoptic to display multiple elements
  * JDrawIn/JDrawOut: used only if JDRAW_HOOK defined, to transform data between synoptic and selection
  * lastWindowClosed: signal will close all opened widgets and dialogs on application exit

h3. Extending Taurus GUI
------------------------

We will differentiate between:

 - Tools (applications launched in a separate process)
 - Panels (widgets embeddable in the application)
 - Applets (widgets or applications launched from the right-side bar)
 - Widgets (widgets not instantiated, but added to the catalog when adding a new panel).
 
The config variables to include new Tools/Panels/Apps/Widgets are:

Use EXTRA_WIDGETS to add new widgets to the catalog and the right-side toolbar.
They will be added to EXTRA_CATALOG_WIDGETS, that contains default extra widgets.

Use EXTRA_PANELS to force new Panels always at startup.
They will be instantiated and running in background.

Use EXTRA_APPS to add launchers to the toolbar.
Creates new AppletDescription objects to add elements to the right-side toolbar.

Use EXTRA_TOOLS to add launchers to the "External Apps" toolbar.
Creates new ExternalApp objects to add elements to the top menu and toolbar.

Use TOOLBARS to add new Toolbars: 
  TOOLBARS = [('Sectors','SR.sector.get_sectors_toolbar')]
Creates new ToolBarDescription objects.


h2. Config Variables
====================

Described below. All of them can be overriden in TangoDB(VACCA free properties) or
using environment variables (adding the VACCA_ preffix).

"""

#==============================================================================
# Import section. You probably want to keep this line. Don't edit this block 
# unless you know what you are doing

import time,os,sys,traceback,imp,inspect
import fandango as fn,taurus
import vacca,vacca.utils
from .doc import get_autodoc,get_vars_docs
from fandango import partial,FakeLogger as FL
from taurus.external.qt import Qt
from taurus.external.qt.Qt import SIGNAL
from taurus.qt.qtgui.taurusgui.utils import \
  PanelDescription, ExternalApp, ToolBarDescription, AppletDescription
from vacca.panel import VaccaAction,VaccaSplash
from vacca.utils import *

# (end of import section)
#==============================================================================

VACCA_CONFIG = os.getenv('VACCA_CONFIG')
PROPS = get_config_properties(VACCA_CONFIG)
VACCA_DIR = WDIR = PROPS.get('VACCA_DIR',VACCA_DIR)

print('-'*80)
print('In vacca.config(done=%s,vacca_dir=%s)'%(globals().get('CONFIG_DONE',False),VACCA_DIR))

try:

    if VACCA_DIR:
        print('Changing directory to VACCA_DIR:%s' %(VACCA_DIR))
        os.chdir(os.environ['VACCA_DIR'])
        sys.path.append(os.environ['VACCA_DIR'])

    try:
        app = Qt.QApplication.instance()
        assert app,'QApplication not running!'
        splash = VaccaSplash()
    except:
        print("QApplication Instance do not exist!")
        app = None

    try:
        import default
    except:
        try:
            default = get_config_file(imp.find_module('vacca')[1]+'/default.py')
        except:
            traceback.print_exc()

    #===============================================================================
    # Loading of Config Files
    #===============================================================================

    #: Config file where perspectives will be saved
    SETTINGS = '/home/$USER/.config/$ORGANIZATION/$GUI_NAME'

    #: Default options are set in default.py file and later customized in a config file.
    #: This CONFIG can be a .py file passed as argument or a TANGO_HOST.py file in
    #: vacca folder.
    CONFIG = get_config_file()

    #: ALL Variables that can be defined in CONFIG FILE
    OPTION_LISTS = [
        ('General Info', ['GUI_NAME','URL_HELP','URL_LOGBOOK','VACCA_LOGO','ORGANIZATION','ORGANIZATION_LOGO',]),
        ('Installation', ['WDIR','SETTINGS']),
        ('Sub systems', ['DOMAIN','TARGET','COMPOSER',]),
        ('Device Tree', ['USE_DEVICE_TREE','CUSTOM_TREE','DEVICES','EXTRA_DEVICES']),
        ('Synoptics', ['JDRAW_FILE','JDRAW_TREE','JDRAW_HOOK','GRID',]),
        ('Device Panel', ['USE_DEVICE_PANEL','DEVICE','PANEL_COMMAND','AttributeFilters','CommandFilters','IconMap',]),
        ('Plot', ['GAUGES']),
        ('Widgets', ['EXTRA_WIDGETS','EXTRA_PANELS','TOOLBARS','MENUS','EXTRA_APPS','EXTRA_CATALOG_WIDGETS']),]

    OPTIONS = [o for c,l in OPTION_LISTS for o in l]
    
    for c,l in OPTION_LISTS:
        __doc__ += get_vars_docs(__name__,title=c+' Options',module_vars=l,subtitle=False)
    
    if CONFIG:
        print('\n%s Config Options:\n'%VACCA_CONFIG)

        #Check renamed variables
        if getattr(CONFIG,'EXTRA_DEVICES',[]) and not getattr(CONFIG,'DEVICES',None):
            CONFIG.DEVICES = CONFIG.EXTRA_DEVICES
        if not getattr(CONFIG,'DEVICE',None) and getattr(CONFIG,'COMPOSER',None):
            CONFIG.DEVICE = CONFIG.COMPOSER

        varpatch = lambda old,new: (
          setattr(CONFIG,new,getattr(CONFIG,old)) 
          if getattr(CONFIG,old,None) and not getattr(CONFIG,new,None) 
          else none)
          
        #Variable replacement (update config ->  update default -> load default)
        for op in OPTIONS:
            limit = 800
            if 'VACCA_'+op in os.environ:
                print('%s.%s overriden by OS'%(VACCA_CONFIG,op))
                setattr(CONFIG,op,os.getenv('VACCA_'+op))
                
            elif op in PROPS:
                print('%s.%s overriden by TangoDB'%(VACCA_CONFIG,op))
                setattr(CONFIG,op,replace_env(PROPS[op]))
                
            if hasattr(CONFIG,op):
                v = replace_env(getattr(CONFIG,op))
                print('\t%s: \t%s = %s'%(CONFIG.__name__,op,str(v).replace('\n',',')[:limit]))
                setattr(default,op,v)
    
    #Adding all variables to Namespace where taurusgui can found them
    try:
        for k,v in vars(default).items():
            if not any((k.startswith('_'),inspect.ismodule(v))): #inspect.isfunction(v)
                vars()[k] = v
    except:
        traceback.print_exc()
    
    #===============================================================================
    # General info.
    #===============================================================================
    #: GUI_NAME will be used on application title and settings filenames
    GUI_NAME = '%s-%s-%s'%(GUI_NAME,VACCA_CONFIG.split('/')[-1].split('.')[0],DB_HOST)
    #: Name to be shown in right-side bar
    ORGANIZATION = ORGANIZATION
    #: Logo to be shown in right-side bar
    ORGANIZATION_LOGO = ORGANIZATION_LOGO
    
    print('\nConfig loaded, Perspectives will be saved/loaded from %s'%(
        SETTINGS.replace('$USER',get_env_variable('USER')
          ).replace('$ORGANIZATION',ORGANIZATION
          ).replace('$GUI_NAME',GUI_NAME)))
    
    SINGLE_INSTANCE = False

    #: CUSTOM_LOGO. It can be an absolute path,or relative to the app dir or a
    #: resource path. If commented out, ":/taurus.png" will be used
    CUSTOM_LOGO = VACCA_LOGO

    #: You can provide an URI for a manual in html format
    #: (comment out or make MANUAL_URI=None to skip creating a Manual panel)
    MANUAL_URI = URL_HELP #'http://packages.python.org/taurus'
    
    print('\n'+'#'*80+'\n')
    print('In vacca.config(%s): loading classes\n'%VACCA_CONFIG)

    #===============================================================================
    # Define panels to be shown.
    # To define a panel, instantiate a PanelDescription object (see documentation
    # for the gblgui_utils module)
    #===============================================================================
    
    print('\t>>> Loading Trend panel ... %s'%','.join(GAUGES))
    
    #: If defined, the default trend will be shown in logarithmic scale
    GAUGES=GAUGES
    if GAUGES:
        trend = PanelDescription('Gauges',
            classname = 'vacca.plot.PressureTrend',
            model = GAUGES)

    trends = PanelDescription('Trends',
        classname = 'vacca.plot.VaccaTrend',
        model = '')

    #: USE_DEVICE_TREE:  True or False, To Show by default the Device_Tree
    USE_DEVICE_TREE = USE_DEVICE_TREE
    #print('\t>>> Parsing DEVICES: %s ...'%(str(DEVICES)[:80]))
    
    try:
        DEVICES = expand_device_list(DEVICES)
        DEVICES.extend(expand_device_list(EXTRA_DEVICES))
        if DEVICE and DEVICE.lower() not in DEVICES:
          DEVICES.append(DEVICE.lower())
        if DEVICES and not DEVICE: 
          DEVICE = DEVICES[0]
        DEVICES = sorted(set(DEVICES))
    except: 
        traceback.print_exc()

    if any((USE_DEVICE_TREE,JDRAW_FILE,len(DEVICES)>1)):
        print('\t>>> Loading Tree panel(%s) ...'%(len(DEVICES)))
        from tree import *
        try:
            # The lastWindowClosed() signal will close all opened widgets and dialogs on application exit
            assert Qt.QApplication.instance(),'QApplication not running!'
            panelclass = VaccaPanel #VaccaAction(default=PANEL_COMMAND)
            Qt.QObject.connect(Qt.QApplication.instance(), Qt.SIGNAL(
             "lastWindowClosed()"), panelclass.kill )
            VaccaTree.setDefaultPanelClass(panelclass)
        except:
            print("Cannot instance Device Tree")
        logger = fn.Logger()
        printf = logger.info

        def filterMatching(a,dct=AttributeFilters,p=printf):
            match = False
            if a.lower().endswith('/state'): return True
            elif a.lower().endswith('/status'): return False
            for k,l in dct.items():
                if fn.searchCl(k,a.rsplit('/',1)[0]):
                    for t in l: #T is every declared Tab for panel (TabName,[attrlist]); or just attrname when not using tabs
                        p((k,t))
                        f = t[-1] if all(map(fn.isSequence,(t,t[-1]))) else [t]
                        if any(fn.matchCl(v,a.rsplit('/',1)[-1]) for v in f): 
                            match =True
            return match
        VaccaTree.setDefaultAttrFilter(filterMatching)
        if IconMap: VaccaTree.setIconMap(IconMap)
        
        tree = PanelDescription('Tree',
            classname = 'vacca.VaccaTree',#'vacca.VaccaTree',#'TaurusDevTree',
            model = CUSTOM_TREE or ','.join(DEVICES),
            sharedDataRead={'LoadItems':'addModels',
                ##DISABLED BECAUSE TRIGGERED RECURSIVE SELECTION, TO BE AVOIDED IN TREE
                #'SelectedInstrument':'findInTree',
                }, #It will load devices from synoptic
            sharedDataWrite={'SelectedInstrument':'deviceSelected(QString)'}
            )

    #: USE_DEVICE_PANEL:  True or False, To Show by default the DevicePanel
    USE_DEVICE_PANEL = USE_DEVICE_PANEL
    if USE_DEVICE_PANEL:
        print('\t>>> Loading Device panel (%s)...' % DEVICE)
        from vacca.panel import VaccaPanel
        m = DEVICE if DEVICE and fn.check_device(DEVICE) else None
        panel = VaccaPanel.getPanelDescription('Device',model=m)
        
    from vacca.panel import VaccaPanel
    if AttributeFilters: VaccaPanel.setAttributeFilters(AttributeFilters)
    if IconMap: VaccaPanel.setIconMap(IconMap)
    if CommandFilters: VaccaPanel.setCommandFilters(CommandFilters)

    #: JDRAW_FILE:  The JDRAW file to create the Synoptic, 
    #: it can be .jdw or .svg but the second option will require the svgsynoptic module from maxlab.
    JDRAW_FILE = JDRAW_FILE

    if JDRAW_FILE:
        print('\t>>> Loading Synoptic panel new ... %s, %s, %s'%(JDRAW_FILE,
                                                         JDRAW_HOOK, JDRAW_TREE))
        from vacca.synoptics import VaccaSynoptic
        try:
            synoptic = VaccaSynoptic.getPanelDescription('Synoptic',JDRAW_FILE,JDRAW_HOOK,JDRAW_TREE)
        except:
            traceback.print_exc()
            sys.exit()

    #: GRID:  True/False to show by default ehe GRID Panel.
    GRID = GRID

    if GRID:
        print('\t>>> Loading Grid panels ...')
        
        GRID['frames'] = False
        GRID['labels'] = False
        GRID['units'] = False
        
        from vacca.grid import VaccaGrid, VaccaVerticalGrid

        try:
            assert Qt.QApplication.instance(),'QApplication not running!'
            grid = VaccaGrid.getGridPanelDescription(GRID)
        except:
            print('\t>>> Unable to create Grid')

        ##VGRID disabled, the Vertical layout should be added as option to common GRID
        #try:
            #assert Qt.QApplication.instance(),'QApplication not running!'
            #vgrid = VaccaVerticalGrid.getVerticalGridPanelDescription(GRID)
        #except:
            #print 'Unable to create VerticalGrid'

    #: COMPOSER:  True/False to show by default the COMPOSER Panel.
    COMPOSER = COMPOSER

    if COMPOSER:
        def VacuumProfile():
            print('\t>>> Loading VacuumProfile()')
            from . import plot
            p = plot.VaccaProfilePlot()
            p.setModel(COMPOSER)
            return p
        try:
            print('\t>>> Loading Profile panel ...')
            profile = PanelDescription('Profile',
                            classname = 'vacca.plot.VaccaProfilePlot',
                            model = COMPOSER,
                            )
        except:
            print('\t>>> Unable to create ProfilePlot')
            
    import vacca.properties
    properties = vacca.properties.VaccaPropTable.getPanelDescription('Properties',model=DEVICE or '')


    #: EXTRA_PANELS:  dictionary of Extra Panels to be shown by default.
    #: Each element may be a PanelDescription object or a tuple with its arguments.
    #:
    #: The dictionary will be defined like:
    #:  * key: Panel name
    #:  * value: list of arguments to PanelDescription
    #:  * value[0]: panel name
    #:  * value[1]: class name
    #:  * value[2]: default model
    #:  * value[3]: shared data read signal
    #:  * value[4]: shared data write signal
    #:
    #: As example, to add PANIC as extra panel use:
    #: EXTRA_PANELS['PANIC'] = ('PANIC','panic.gui.AlarmGUI','',{},{'HighlightInstruments':'devicesSelected'})
    EXTRA_PANELS = EXTRA_PANELS

    if EXTRA_PANELS:
        print('\t>>> Loading Extra panels ... %s'%str(EXTRA_PANELS))
        if not hasattr(EXTRA_PANELS,'items'):
            EXTRA_PANELS = dict(('extra%d'%(i+1),p) for i,p in enumerate(EXTRA_PANELS))
            
        def get_panel(i):
            pargs = EXTRA_PANELS[i]
            if not fn.isSequence(pargs): return pargs
            elif len(pargs)==3: return PanelDescription(pargs[0],classname=pargs[1],model=pargs[2])
            else: return PanelDescription(pargs[0],classname=pargs[1],model=pargs[2],sharedDataRead=pargs[3],sharedDataWrite=pargs[4])

        for k,p in EXTRA_PANELS.items():
            try:
                vars()[k] = get_panel(k)
            except:
                traceback.print_exc()

    #===============================================================================
    #: EXTRA_WIDGETS: The Dictionary of EXTRA_WIDGETS Panels
    #:
    #: Adding other widgets to the catalog of the "new panel" dialog.
    #: pass a tuple of (classname,screenshot)
    #:
    #: - classname may contain the module name.
    #: - screenshot can either be a file name relative to the application dir or
    #:   a resource URL or None
    #:
    #: examples:
    #:
    #:    ('vacca.properties.VaccaPropTable',wdir('vacca/image/widgets/Properties.png')),
    #:    ('vacca.panel.VaccaPanel',wdir('vacca/image/widgets/Panel.png')),
    EXTRA_WIDGETS = EXTRA_WIDGETS

    #: EXTRA_CATALOG_WIDGETS: The Dictionary of EXTRA_CATALOG_WIDGETS Panels to Show by
    #: default.
    #:Is the Sum of EXTRA_WIDGETS and the custom Widgets defined in config.py
    EXTRA_CATALOG_WIDGETS = []
    EXTRA_CATALOG_WIDGETS = EXTRA_WIDGETS+[
        #('vacca.VacuumProfile',WDIR+'image/ProfilePlot.jpg'),
        #('vacca.plot.PressureTrend',WDIR+'image/PressureTrend.jpg'),
        #('vacca.VacuumGrid',WDIR+'image/BLGrid.jpg'),
        #('vacca.VerticalGrid',WDIR+'image/VerticalGrid.jpg'),
        ('vacca.properties.VaccaPropTable',vpath('image/widgets/Properties.png')),
        ('fandango.qt.QEvaluator',':/snapshot/large/snapshot/TaurusShell.png'),
        ]

    #===============================================================================
    #: TOOLBARS
    #:
    #: Define custom toolbars to be shown. To define a toolbar, instantiate a
    #: ToolbarDescription object (see documentation for the gblgui_utils module)
    #===============================================================================
    
    #dummytoolbar = ToolBarDescription('Empty Toolbar',
                            #classname = 'QToolBar',
                            #modulename = 'PyQt4.Qt')
    
    #panictoolbar = ToolBarDescription('Panic Toolbar',
    #                        classname = 'PanicToolbar',
    #                        modulename = 'tangopanic')
    
    toolbars = []
    for name,obj in (TOOLBARS or []):
        toolbars.append(ToolBarDescription(name,
            classname = obj.split('.')[-1],
            modulename = obj.rsplit('.',1)[0],
            #sharedDataWrite={'selectedPerspective':'blabla'}
            ))
        toolbar = toolbars[-1]
        
    try:
        for menu,actions in MENUS:
            vacca.utils.add_menu(menu,actions)
    except:
        traceback.print_exc()
        
    #===============================================================================
    # Define custom applets to be shown in the applets bar (the wide bar that
    # contains the logos). To define an applet, instantiate an AppletDescription
    # object (see documentation for the gblgui_utils module)
    #===============================================================================
    
    # ALREADY LOADED FROM vacca.default.EXTRA_APPS
    
    #xmambo = AppletDescription('Mambo',classname = 'vacca.panel.VaccaAction',
    # model=["Archiving",wdir('vacca/image/widgets/ProfilePlot.png'),'mambo'],)
    
    #===============================================================================
    # Define which External Applications are to be inserted.
    # To define an external application, instantiate an ExternalApp object
    # See TaurusMainWindow.addExternalAppLauncher for valid values of ExternalApp
    #===============================================================================

    # ALREADY LOADED FROM vacca.default.EXTRA_TOOLS
    
    #xvacca = ExternalApp(cmdargs=['konqueror',URL_HELP], text="Alba VACuum Controls Application", icon=WDIR+'image/icons/cow-tux.png')
    #DEFAULT_APPS = 
    
    EXTRA_TOOLS = EXTRA_TOOLS
    for name,cmdargs,icon in EXTRA_TOOLS:
        locals()['x'+name] = ExternalApp(cmdargs=cmdargs,text=name,icon=icon or None)
    
    #xjive = ExternalApp(cmdargs=['jive'], text="Jive")#, icon=WDIR+'image/icons/cow-tux.png')
    #xastor = ExternalApp(cmdargs=['astor'], text="Astor")#, icon=WDIR+'image/icons/cow-tux.png')
    
    #===============================================================================
    # POOL RELATED OPTIONS
    #===============================================================================
    
    # Set INSTRUMENTS_FROM_POOL to True for enabling auto-creation of
    # instrument panels based on the Pool Instrument info
    #MACROSERVER_NAME = 
    #DOOR_NAME = 
    #MACROEDITORS_PATH = 
    INSTRUMENTS_FROM_POOL = False
    
    #===============================================================================
    # THIS SYNOPTIC OPTION IS POOL-related; IT'S NOT THE MAIN APPLICATION SYNOPTIC!!!
    # If you want an instrument selection synoptic, set the SYNOPTIC variable
    # to the file name of a jdraw file. If a relative path is given, the directory
    # containing this configuration file will be used as root
    # (comment out or make SYNOPTIC=None to skip creating a synoptic panel)
    #===============================================================================
    SYNOPTIC = [] #'images/example01.jdw','images/syn2.jdw']    
    
    #===============================================================================

    #: 
    
    #Forcing nesting of dock widgets
    if app:
        try:
            main = vacca.utils.get_main_window(app)
            main.setDockNestingEnabled(True)
        except:
            traceback.print_exc()
            
    EXTRA_APPS = EXTRA_APPS
    """ The Vacca Panels to show in the JogsBar. Use EXTRA_APPS to add launchers to the toolbar.
    It will create new AppletDescription objects to add elements to the right-side toolbar.

    Examples:
     * EXTRA_APPS['Properties'] = {'class' : vacca.VaccaPropTable}
     * EXTRA_APPS['DevicePanel'] = {'class' : vacca.VaccaPanel}
     * EXTRA_APPS['Panic']= {'class' : vacca.VaccaPanic       }
     * EXTRA_APPS['ExtraDock']= {'class' : Qt.QMainWindow       }
    """
    adder = vacca.addCustomPanel2Gui(EXTRA_APPS)            

    print('\n\t>>> Config Finished ...')
    globals()['CONFIG_DONE'] = True
    print('-'*80 + '\n\n\n')

except:
    print(traceback.format_exc())




#!/usr/bin/env python

#############################################################################
##
## This file is part of Taurus, a Tango User Interface Library
## 
## http://www.tango-controls.org/static/taurus/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
## 
## Taurus is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
## 
## Taurus is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
## 
## You should have received a copy of the GNU Lesser General Public License
## along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
###########################################################################

#print '#'*80+'\n'+'vacca.__init__'

__doc__ = """
-------------------------------------------------------------
VACCA, the Viewer and Commander Control Application for TANGO
-------------------------------------------------------------

The *vacca* library is divided in different submodules:

 * config : the vacca MAIN file, a config setup to highly customize TaurusGUI instances
 * doc : helper methods to build documentation
 * setup :  script to download and install vacca and its dependencies from sourceforge
 * utils : miscellaneous methods to manage folders, config files, look&feel 

Several other modules contain the widget catalog: 

 * alarms
 * grid : grid-like synoptics
 * panel : device/attribute panels and forms
 * plot : trends and plots
 * properties : property editor widget
 * synoptics
 
"""
import traceback

try:
  from vacca.doc import get_autodoc
except Exception,e: print('vacca.doc not loadable')
from vacca.panel import *
from vacca.tree import *
from vacca.utils import *
from vacca.properties import *
from vacca.grid import *
try:
  from vacca.alarms import *
except Exception,e: print('vacca.alarms not loadable')

## THIS DIDNT WORKED!!
#print('Autogenerating VACCA docs')
#import panel,tree,utils,properties,grid,vaccaPanic,doc
#for m in panel,tree,utils,properties,grid,vaccaPanic:
    #print m.__name__
    #if not m.__doc__:
        #m.__doc__ = "\n%s\n%s\n\n" % (m.__name__,'='*len(m.__name__))
    #if ".. auto" not in m.__doc__:
        #m.__doc__+=doc.get_function_docs(m.__name__,vars(m))
        #m.__doc__+=doc.get_class_docs(m.__name__,vars(m))

__doc__+="\nextended"

#try:
 #from config import *
#except Exception,e:
 #print 'Unable to import vacca.config: %s'%str(e)



#!/usr/bin/env python

#############################################################################
##
## This file is part of Taurus, a Tango User Interface Library
## 
## http://www.tango-controls.org/static/taurus/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
## 
## Taurus is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
## 
## Taurus is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
## 
## You should have received a copy of the GNU Lesser General Public License
## along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

import traceback,time
import fandango
import PyTango
import taurus
from vacca.utils import Qt,Qwt5
from taurus.qt.qtgui.plot import TaurusTrend,TaurusPlot
from taurus.qt.qtgui.panel import TaurusDevicePanel
from fandango.qt import Draggable

try:
  from PyTangoArchiving.widget.trend import ArchivingTrend
  TREND_CLASS = TaurusTrend  #TaurusArchivingTrend
except:
  TREND_CLASS = TaurusTrend

__doc__ = """
Plots in VACCA have several setups:

 * ApplyConfig is disabled, to avoid previous configs to override trend behavior
 * Default Trend length has several hours instead of seconds.
 * Archiving is still disabled, but also Scale change warnings when re-enabled.
 
"""

class PressureTrend(TREND_CLASS):
        
    def showEvent(self, event):
        if not getattr(self, '_tuned', False):
            setup_pressure_trend(self)
            setattr(self, '_tuned', True)
        TREND_CLASS.showEvent(self, event)
        
class VaccaTrend(TREND_CLASS):
    
    def showEvent(self, event):
        if not getattr(self, '_tuned', False):
            setup_pressure_trend(self,log=False,length=4*3600)
            setattr(self, '_tuned', True)
        TREND_CLASS.showEvent(self, event)

def setup_pressure_trend(tt,log=True,length=12*3600):
    print '*'*80
    print 'in setup_pressure_trend(%s,length=%s s)' % (tt, length)
    from taurus.external.qt import Qwt5
    try:
        #n,w = str(qtw.tabText(qtw.count()-1)),qtw.widget(qtw.count()-1)
        #qtw.insertTab(0,w,n)
        tt.setXDynScale(True)
        xMax = time.time() #tt.axisScaleDiv(Qwt5.QwtPlot.xBottom).upperBound()
        #xMin = tt.axisScaleDiv(Qwt5.QwtPlot.xBottom).lowerBound()
        rg = length #abs(self.str2deltatime(str(self.ui.xRangeCB.currentText())))
        xMin = xMax-rg
        tt.setAxisScale(Qwt5.QwtPlot.xBottom, xMin, xMax)
        if log:
            tt.setAxisScaleType(Qwt5.QwtPlot.yLeft,
                Qwt5.QwtScaleTransformation.Log10)
            #tt.setAxisScale(Qwt5.QwtPlot.yLeft,1e-11,1e-2)
        tt.setUseArchiving(False)
        tt.setModelInConfig(False)
        tt.disconnect(tt.axisWidget(tt.xBottom), 
            Qt.SIGNAL("scaleDivChanged ()"),tt._scaleChangeWarning)
        #tt.setCurvesYAxis([tcs],tt.yRight)
        #: Disabling loading of configuration from file; to avoid a faulty setup to continuously crash the application.
        setattr(tt, 'applyConfig', (lambda *k, **kw: None))
    except:
        print 'Exception in set_pressure_trend(%s)' % tt
        print traceback.format_exc()
    print '*'*80

class VaccaProfilePlot(Draggable(TaurusPlot)):
    
    #def getModelClass(self):
        #return taurus.core.TaurusDevice
    def __init__(self, *args, **kwargs):
        TaurusPlot.__init__(self, *args, **kwargs)
        self._profile_loaded = False
        self._positions = []
        self._labels = []
        self.setModelInConfig(False)
        
    def setModel(self, model):
        print '*'*80
        self.info('VaccaProfilePlot.setModel(%s)' % model)
        print '*'*80
        try:
            #if self._profile_loaded: return
            if fandango.isSequence(model) or 'attributename' in fandango.tango.parse_tango_model(model):
                self.info('setting an attribute model')
                TaurusPlot.setModel(self, model)# model = model[0]# str(
                # model).rsplit('/',1)[0]
            else:
                self.info('setting a composer model')
                assert fandango.check_device(model)
                dev = taurus.Device(model)
                if all(a in map(str.lower, dev.get_attribute_list()) for a in
                    ('ccgaxxis', 'ccgpressures', 'ipaxxis', 'ippressures',
                    'thermoaxxis', 'thermocouples', 'axxispositions',
                     'axxislabels')):
                    TaurusPlot.setModel(self, [])
                    setup_profile_plot(self, model)
                else:
                    self.warning('%s has not all required attributes' % model)
            if len(self._positions) and len(self._labels):
                self.info('Setting CustomXLabels ...')
                self.setAxisCustomLabels(Qwt5.QwtPlot.xBottom,
                                         zip(self._positions, self._labels), 60)
        except Exception, e:
            self.warning('VaccaProfilePlot.setModel(%s) failed!: %s' % (model, e))
            
def setup_profile_plot(tp, composer, picker=True):
    #,gauges,pumps,tcs,positions=None,labels=None):
    """ This method configures a tau plot to show pressure/temperature profiles """
    tp.info('setup_profile_plot(%s,%s)' % (composer, picker))
    #tp.setModel('BO/VC/Composer/IPAxxis|BO/VC/Composer/IPPressures')
    #tp.setModel('BO/VC/Composer/CCGAxxis|BO/VC/Composer/CCGPressures,BO/VC/Composer/IPPressures')
    #positions = [0,17,31,33,49,63,80,94,111,114,116,122]
    #labels=['Q1A','Q1B','Q1LTB','Q2A','Q2B','Q3A','Q3B','Q4A','Q4B','Q4BT','Q4C','Q4D+RF']
    gauges = '%s/CCGAxxis|%s/CCGPressures' % (composer, composer)
    pumps = '%s/IPAxxis|%s/IPPressures' % (composer, composer)
    tcs = '%s/ThermoAxxis|%s/Thermocouples' % (composer, composer)
    positions = '%s/AxxisPositions' % composer
    labels = '%s/AxxisLabels' % composer
                
    pumps, gauges, tcs = [str(s).lower() for s in (pumps, gauges, tcs)]
    tp.setModelInConfig(False)
    tp.setModifiableByUser(False)
    tp.setModel(pumps)
    tp.addModels([tcs])
    tcs = tcs.split('|')[-1]
    tp.setCurvesYAxis([tcs], tp.yRight)
    tp.addModels([gauges])
    gauges = gauges.split('|')[-1]
    Curves = {}
    for cname in (pumps, gauges, tcs):
        tp.info('set curve %s' % cname)
        try:
            Curves[cname] = tp.curves[cname]
        except:
            Curves[cname] = tp.curves[cname.split('|')[-1]]
        
    if positions and labels:
        tp.info('Setting CustomXLabels ...')
        try:
            tp._positions = list(int(i) for i in taurus.Attribute(positions).read().value)
            tp._labels = list(str(i) for i in taurus.Attribute(labels).read().value)
            tp.setAxisCustomLabels(Qwt5.QwtPlot.xBottom,zip(tp._positions,
                                                            tp._labels), 60)
        except:
            print 'Unable to read pressure profile labels'
            tp._positions = []
            tp._labels = []
            #print traceback.format_exc()
    
    for cname, color, width in [(gauges, Qt.Qt.blue, 3), (pumps,
                                                          Qt.Qt.magenta, 1),
                                (tcs, Qt.Qt.red, 2)]:
        cap = Curves[cname].getAppearanceProperties()
        cap.lColor = color
        cap.lWidth = width
        tp.setCurveAppearanceProperties({Curves[cname].getCurveName(): cap})
        
    Curves[gauges].setSymbol(Qwt5.QwtSymbol(Qwt5.QwtSymbol.Diamond,
                                            Qt.QBrush(Qt.Qt.yellow),
                                            Qt.QPen(Qt.Qt.green),
                                            Qt.QSize(7, 7)
                                            )
                             )
    #f1 = Qwt5.QwtSplineCurveFitter()
    #f1.setSplineSize(2000)
    #Curves[pumps].setCurveFitter(f1)
    #Curves[pumps].setCurveAttribute(Curves[pumps].Fitted,True)
    tp.enableAxis(tp.yRight)        
    tp.setAxisScaleType(Qwt5.QwtPlot.yLeft, Qwt5.QwtScaleTransformation.Log10)
    #tp.setAxisScale(Qwt5.QwtPlot.yLeft,1e-11,1e-4)
    tp.setCanvasBackground(Qt.Qt.white)
    tp.toggleDataInspectorMode(True)
    tp.showMaxPeaks(True)
    Curves[pumps].showMaxPeak(False)
    tp.setAllowZoomers(False)
    tp.toggleDataInspectorMode(True)
    
    if picker:
        tp.info('setting Picker ...')
        tp.disconnect(tp._pointPicker, Qt.SIGNAL('selected(QwtPolygon)'),
                      tp.pickDataPoint)
        tp.connect(
            tp._pointPicker, 
            Qt.SIGNAL('selected(QwtPolygon)'), 
            lambda pos, s = tp: pickPlotPoint(s, pos, xlabels=(zip(
                tp._positions, tp._labels) if tp._positions and tp._labels
                                                               else None))
            )

    tp._profile_loaded = True
    return tp    
    
def pickPlotPoint(self, pos, scope=20, showMarker=True,
                  targetCurveNames=None, xlabels=None):
    '''Finds the pyxel-wise closest data point to the given position. The
    valid search space is constrained by the scope and targetCurveNames
    parameters.

    :param pos: (Qt.QPoint or Qt.QPolygon) the position around which to look
                for a data point. The position should be passed as a
                Qt.QPoint (if a Qt.QPolygon is given, the first point of the
                polygon is used). The position is expected in pixel units,
                with (0,0) being the top-left corner of the plot
                canvas.

    :param scope: (int) defines the area around the given position to be
                    considered when searching for data points. A data point is
                    considered within scope if its manhattan distance to
                    position (in pixels) is less than the value of the scope
                    parameter. (default=20)

    :param showMarker: (bool) If True, a marker will be put on the picked
                        data point. (default=True)

    :param targetCurveNames: (sequence<str>) the names of the curves to be
                                searched. If None passed, all curves will be
                                searched

    :return: (tuple<Qt.QPointF,str,int> or tuple<None,None,None>) if a point
                was picked within the scope, it returns a tuple containing the
                picked point (as a Qt.QPointF), the curve name and the index of
                the picked point in the curve data. If no point was found
                within the scope, it returns None,None,None
    '''
    print "pickPlotPoint(...)"
    if isinstance(pos, Qt.QPolygon):
        pos = pos.first()
    scopeRect = Qt.QRect(0, 0, scope, scope)
    scopeRect.moveCenter(pos)
    mindist = scope
    picked = None
    pickedCurveName = None
    pickedIndex = None
    self.curves_lock.acquire()
    try:
        if targetCurveNames is None:
            targetCurveNames = self.curves.keys()
        #print '%d targetCurveNames'%len(targetCurveNames)
        for name in targetCurveNames:
            curve = self.curves.get(name, None)
            if curve is None: 
                #print("Curve '%s' not found"%name)
                continue
            if not curve.isVisible(): 
                #print("Curve '%s' not visible"%name)
                continue
            data=curve.data()
            #print("Curve '%s' has %d points"%(name,data.size()))
            for i in xrange(data.size()):
                point = Qt.QPoint(self.transform(curve.xAxis(), data.x(i)),
                                 self.transform(curve.yAxis(), data.y(i))
                                  )
                if scopeRect.contains(point):
                    #print( 'Comparing %s,%s vs %s'%(pos,scopeRect,point))
                    dist = (pos-point).manhattanLength()
                    if dist < mindist:
                        mindist = dist
                        picked = Qt.QPointF(data.x(i), data.y(i))
                        pickedCurveName = name
                        pickedIndex = i
                        pickedAxes = curve.xAxis(), curve.yAxis()
                        found = True
            #if picked: print("Curve '%s' contains %s"%(name,pos))
    finally:
        self.curves_lock.release()
    
    if picked is None:
        print 'pickPlotPoint(%s)> No matching point found for any curve' % pos
        return
    
    xlabels = xlabels or []
    print("pickPlotPoint(x=%s,y=%s,xlabels=[%s])" % (picked.x(), picked.y(),
                                                     len(xlabels)
                                                     ))

    if showMarker and picked is not None:
        print 'showing pickedMarker'
        self._pickedMarker.detach()
        self._pickedMarker.setValue(picked)
        self._pickedMarker.setAxis(*pickedAxes)
        self._pickedMarker.attach(self)
        self._pickedCurveName = pickedCurveName
        self._pickedMarker.pickedIndex = pickedIndex
        self.replot()
        label = self._pickedMarker.label()
        try: 
            xi, xl = ([(x[0], x[1]) for x in (xlabels or []) if x[
                0] <= picked.x()] or [[-1, None]])[-1]
            c = self.curves[pickedCurveName]
            #xc = [self.transform(c.xAxis(),x) for x in c.data().xData()]
            xc = [x for x in c.data().xData()]
            xx = len([x for x in xc if x >= xi and x <= picked.x()])
            #print [x[0] for x in xlabels]
            #print 'picked point: %s'%picked.x()
            #print 'last label %s at %s'%(xl,xi)
            #print 'data point are %s'%(xc)
            #print 'data points between pick and label: %s'%xx
            index = xx
            tag = '%s-%02d' % (xl, index)
            if xl and '/ip' in pickedCurveName.lower():
                try:
                    tag = PyTango.AttributeProxy(tag.replace('-', '/VC/IP-')
                                                 +'/Controller').read().value
                except:
                    import traceback
                    print traceback.format_exc()
        except: 
            import traceback
            print traceback.format_exc()
            index, xl = (-1, None)
            tag = ''
        if xl:
            infotxt = "'%s'[%i]:\n\t (%s,%.2e)" % (pickedCurveName,
                                                  pickedIndex, tag, picked.y())
        elif self.getXIsTime():
            infotxt = "'%s'[%i]:\n\t (x=%s,%.3g)"%(pickedCurveName,
                                                   pickedIndex,
                                                   datetime.fromtimestamp(
                                                       picked.x()).ctime(),
                                                   picked.y()
                                                   )
        else:
            infotxt = "'%s'[%i]:\n\t (x=%.3g, y=%.3g)" % (pickedCurveName,
                                                         pickedIndex,
                                                          picked.x(),
                                                          picked.y())
        label.setText(infotxt)
        print '\t%s' % infotxt
        fits = label.textSize().width() < self.size().width()
        if fits:
            label.setText(infotxt)
            self._pickedMarker.setLabel(Qwt5.QwtText (label))
            self._pickedMarker.alignLabel()
            self.replot()
        else:
            popup = Qt.QWidget(self, Qt.Qt.Popup)
            popup.setLayout(Qt.QVBoxLayout())
            popup.layout().addWidget(Qt.QLabel(infotxt))  #@todo: make the widget background semitransparent green!
            popup.setWindowOpacity(self._pickedMarker.labelOpacity)
            popup.show()
            popup.move(self.pos().x()-popup.size().width(), self.pos().y())
            popup.move(self.pos())
            Qt.QTimer.singleShot(5000, popup.hide)
        
    return picked,pickedCurveName,pickedIndex

import doc
__doc__ = doc.get_autodoc(__name__,vars())

if __name__ == '__main__':
    import taurus.qt.qtgui.application
    app = taurus.qt.qtgui.application.TaurusApplication()
    cmps = fandango.get_matching_devices('*/vc/all')
    tp = VaccaProfilePlot()
    tp.setWindowTitle(cmps[0])
    tp.setModel(cmps[0])
    tp.show()
    app.exec_()



#!/usr/bin/env python
"""
Vacca runner; this file emulates this call:
 >taurusgui vacca
 
Config file (or Property name) is obtained from shell args, then env, 
then properties in this order.

If empty, a DEFAULT profile is created pointing to default.py
 
MAIN CODE FOR PANELS GENERATION IS IN vacca.config SUBMODULE
"""

import sys,os,re,time,imp,traceback

args = sys.argv[1:]

__doc__ = """

The file vacca/main.py is the vaccagui launcher 

It creates the taurusgui environment and sets all environment variables.

vaccagui usage
--------------

Launching vacca, loading configuration from a target.py:

  > vaccagui [target.py or $CONFIG] [$OPTION=...]
  
  $CONFIG will load values from VACCA.$CONFIG property in the database
  
  $OPTION=... can be used to override values declared properties or target.py
  
Environment variables (optional, will be initialized in vacca/main.py):

  VACCA_CONFIG : if set, equivalent to passing target.py as argument
  VACCA_DIR : directory to resources needed by target.py (target.py folder by default)
  VACCA_PATH : path to vacca module (initialized by imp.find_module())
  
  If not set, default values are those set as VACCA properties in Tango DB.

Reset of QSettings files:

  > vaccagui --reset   #The last saved perspective will be removed    
  > vaccagui --clean   #All the .ini files will be removed
  
Other options:
  > vaccagui --helps   #Prints this text
  > vaccagui --list    #Prints available configurations
    
"""

if args and args[0].strip('- ') == 'help':
  print(__doc__)
  sys.exit(0)
  
def remove_last_config(filename):
    print('vacca.remove_last_config(%s)'%filename)
    lines = open(filename).readlines()
    sections = dict((l.strip(),i) for i,l in enumerate(lines) 
                    if re.match('[\[][a-zA-Z]*[\]]',l))
    sections['[End]'] = len(lines)
    begin = sections['[General]']
    end = min(i for i in sections.values() if i>begin)
    fo = open(filename,'w')
    fo.writelines(lines[:begin]+lines[end:])
    fo.close()  

folder = os.getenv('HOME')+'/.config/VACCA/'
if '--clean' in args:
  print('Removing last vacca configs (%s/*.ini)'%folder)
  os.remove(folder+'*.ini')
  if len(args)==1: sys.exit(0)
  
elif '--reset' in args:
  inits = [a for a in os.walk(f).next()[2] if a.endswith('.ini')]
  print('Removing last vacca configs (%s)'%inits)
  [remove_last_config(folder+filename) for filename in inits]
  if len(args)==1: sys.exit(0)
  
elif '--help' in args or '-h' in args or '-?' in args:
  print(__doc__)
  sys.exit(0)
  
elif '--list' in args:
  import vacca.utils as vu
  configs = vu.get_config_properties()
  print('\nVaccaConfigs:')
  print('\n\t'+'\n\t'.join(configs)+'\n')
  sys.exit(0)
  
###############################################################################
# Delayed imports

import taurus
from taurus.core.util import argparse
from taurus.qt.qtgui.application import TaurusApplication
from taurus.qt.qtgui.taurusgui import TaurusGui

import vacca.utils as vu
import vacca 

###############################################################################

print('-'*80)
print("In vacca.main(%s) ..."%args)

options = [a for a in args if a.startswith('-')]
values = [a for a in args if '=' in a]
files = [a for a in args if a not in options+values] or [os.getenv('VACCA_CONFIG')]
configs = vu.get_config_properties() or vu.create_config_properties()

if not files or not files[0]: files = [configs.keys()[0]]

dirname = os.getenv('VACCA_DIR') or ''
config = os.getenv('VACCA_CONFIG') or ''

if files[0] in configs:
    print('Loading %s'%files[0])
    data = vu.get_config_properties(files[0])
    config = config or data.get('VACCA_CONFIG',files[0])
    dirname = dirname or data.get('VACCA_DIR',dirname)
else: 
    config = config or files[0]
    
if os.path.isfile(config):
  config = os.path.abspath(config)
    
elif config:
    try:
      import imp
      print('Loading %s as python module'%config)
      config = imp.find_module(config.replace('.','/'))[1]
      dirname = os.path.dirname(config)
    except:
      pass

dirname = dirname or os.path.dirname(config) or \
  vu.get_vacca_property('VACCA_DIR',extract=1) or ''

vu.VACCA_DIR = os.environ['VACCA_DIR'] = dirname
vu.VACCA_CONFIG = os.environ['VACCA_CONFIG'] = config

print('Vacca Environment variables (vacca.main):')
print('\n'.join(map(str,(t for t in os.environ.items() if 'VACCA' in t[0]))))
  
### MAIN CODE FOR PANELS GENERATION IS IN vacca.config SUBMODULE
print '-'*80
confname = 'vaccagui'
app = TaurusApplication()
gui = TaurusGui(None, confname=confname)
gui.show()
ret = app.exec_()

taurus.info('Finished execution of TaurusGui')
sys.exit(ret)




#!/usr/bin/env python

#############################################################################
##
## This file is part of Taurus, a Tango User Interface Library
## 
## http://www.tango-controls.org/static/taurus/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
## 
## Taurus is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
## 
## Taurus is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
## 
## You should have received a copy of the GNU Lesser General Public License
## along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

import fandango,vacca,traceback,taurus
from taurus.qt.qtgui.table.taurusdevicepropertytable import TaurusPropTable

from fandango.qt import DoubleClickable,Dropable
from taurus.external.qt import Qt
from taurus.qt.qtcore.mimetypes import TAURUS_ATTR_MIME_TYPE, TAURUS_DEV_MIME_TYPE, TAURUS_MODEL_MIME_TYPE
from taurus.qt.qtgui.taurusgui.utils import PanelDescription

try:
    #Taurus 3
    from taurus.core import TaurusDevice,TaurusAttribute,TaurusDatabase as TaurusAuthority
except:
    #Taurus 4
    from taurus.core import TaurusDevice, TaurusAttribute, TaurusAuthority

class VaccaPropTable(DoubleClickable(Dropable(TaurusPropTable))):
    """
    It is a class that inherits from TaurusPropTable, DoubleClickable (fandango) and Dropable (fandango).

    Shows the Devices Properties, this widget is listening to shareDataManager if any Device has been selected to show its properties

    This class has the follow functionalities:
        * Is connected to shareDataManager to share information in the GUI.
        * Can be edit the property by doubleclick function.
    """

    def __init__(self, parent=None, designMode = False):
        """
        In Init, the class VaccaPropTable check if exist any shareDataManager to
        subscribe in it.

        :param parent:
        :param designMode:
        :return:
        """
        #super(type(self),self).__init__(parent,designMode)
        TaurusPropTable.__init__(self,parent,designMode)
        sdm = vacca.utils.get_shared_data_manager()
        if sdm:
            sdm.connectReader('SelectedInstrument',self.setTable,readOnConnect=True)
        print sorted(a for a in dir(self) if 'check'  in a )
        if self.checkDropSupport():
            self.setSupportedMimeTypes([self.TAURUS_DEV_MIME_TYPE,
                                        self.TAURUS_MODEL_MIME_TYPE,
                                        self.TAURUS_MODEL_MIME_TYPE,
                                        self.TAURUS_MODEL_LIST_MIME_TYPE,
                                        self.TEXT_MIME_TYPE])


        self.setAcceptDrops(True)
        #self.setModifiableByUser(True)
        self.setDropEventCallback(self.setTable)
        self.setModelInConfig(False)
        self.setClickHook(self.editProperty)
        
    @staticmethod
    def getPanelDescription(name='TangoDeviceProperties',model=''):
        """
        :param name: Name for the Panel
        :param model: Model for the panel
        :return:
        """
        return PanelDescription(name,classname='vacca.properties.VaccaPropTable',model=model)
    
    def getModelClass(self):
        return TaurusDevice

    def setModel(self,model): 
        """
        Set Model is the callback used in shareDataManager to manage device
        selections.
        :param model:
        :return:
        """
        try:
          self.setTable(model)
        except:
          traceback.print_exc()
        
    def setTable(self,model,filters=[]):
        ''' 
        This method  overrides TaurusPropTable.setTable(), which connects with TaurusClassTable
        This method fill the table with the names of properties and values for the device selected
        '''      
        try:
            model = model and fandango.tango.parse_tango_model(str(model))
            if model is None:
              self.warning('VaccaPropTable.setTable(%s(%s)): MODEL NOT PARSABLE!'%(type(model),model))
              return
            else:
              try: model = model['device']
              except: model = str(model)
              self.debug('VaccaPropTable.setTable(%s(%s))'%(type(model),model))
            
            #TaurusPropTable.setTable(self,model)
            Qt.QObject.disconnect(self,Qt.SIGNAL("cellChanged(int,int)"),self.valueChanged)
            self.db = fandango.get_database()
            dev_name = str(model)
            self.list_prop = list(self.db.get_device_property_list(dev_name,'*'))
            neg = ['polled_attr']+[f[1:] for f in filters if f.startswith('!')]
            pos = [f for f in filters if not f.startswith('!')]
            self.list_prop = [p for p in self.list_prop if
              (not pos or fandango.matchAny(pos,p)) 
              and not fandango.matchAny(neg,p)]
            
            self.setRowCount(len(self.list_prop))
            for i in range(0,len(self.list_prop)):
                elem = self.list_prop[i]
                self.setText(elem,i,0)
                self.dictionary=self.db.get_device_property(dev_name,self.list_prop)
                self.debug('Getting %s properties: %s -> %s'%(dev_name,self.list_prop,self.dictionary))
                value=self.dictionary[elem]
                self.debug('VaccaPropTable: property %s is type %s'%(elem,type(value)))
                USE_TABLES=False
                if USE_TABLES: self.setPropertyValue(value,i,1)
                else:
                    if not isinstance(value,str): #not something like an string
                        #if isinstance(value,list):#type(value) is list: 
                        heigh1 = len(value)
                        value = '\n'.join(str(v) for v in value) # adding new lines in between elements in the list
                    self.setText(str(value),i,1)
            
            self.updateStyle()
            self.dev_name = dev_name
            #self.dev_obj = taurus.Device(dev_name)
            self.setWindowTitle('%s Properties'%dev_name)
            self.resizeColumnsToContents()
            self.resizeRowsToContents()
            
        except:
            traceback.print_exc()
            
    def contextMenuEvent(self,event):
        ''' This function is called when right clicking on qwt plot area. A pop up menu will be
        shown with the available options. '''
        self.info('TaurusPropTable.contextMenuEvent()')
        menu = Qt.QMenu(self)
        configDialogAction = menu.addAction("Add new property")
        self.connect(configDialogAction, Qt.SIGNAL("triggered()"), self.addProperty)
        configDialogAction = menu.addAction("Delete property")
        self.connect(configDialogAction, Qt.SIGNAL("triggered()"), self.deleteProperty)
        configDialogAction = menu.addAction("Edit property")
        self.connect(configDialogAction, Qt.SIGNAL("triggered()"), self.editProperty)
        menu.addSeparator()
        cmd = 'updateDynamicAttributes' if 'DynamicAttributes' in self.list_prop else 'init'
        configDialogAction = menu.addAction("Execute "+cmd+"()")
        self.connect(configDialogAction, Qt.SIGNAL("triggered()"), 
            lambda d=self.dev_name,c=cmd:
              vacca.utils.YesNoDialog('Warning',
              'Are you sure that %s supports %s() command?'%(self.dev_name,c))
              and taurus.Device(d).command_inout(c))
        menu.exec_(event.globalPos())
        del menu    

    @staticmethod
    def getDefaultIcon():
        """
        :return: The Default Icon Path.
        """
        import vacca.utils
        path = 'image/widgets/Properties.png'
        return vacca.utils.vpath(path)

    @classmethod
    def __test__(klass,arg=None):
        qapp = Qt.QApplication([])
        i = klass()
        print i
        print klass
        i.show()
        if arg: i.setTable(arg)
        qapp.exec_()

__doc__ = vacca.get_autodoc(__name__,vars())

if __name__ == '__main__':
    import sys
    VaccaPropTable.__test__(*sys.argv[1:])
        



#!/usr/bin/env python

#############################################################################
##
## This file is part of Taurus, a Tango User Interface Library
## 
## http://www.tango-controls.org/static/taurus/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
## 
## Taurus is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
## 
## Taurus is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
## 
## You should have received a copy of the GNU Lesser General Public License
## along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

import os,fandango,vacca,traceback
from vacca.utils import vpath,wdir
#from panic.gui import AlarmGUI
from taurus.qt.qtgui.taurusgui.utils import PanelDescription

try:
  from panic.gui.gui import AlarmGUI
except:
  from PanicGUI import AlarmGUI



class VaccaPanic(AlarmGUI):
    """
    It is a class that inherits from AlarmGUI.

    This Panel show the alarms, and allow to interact with them.

    """

    @staticmethod
    def getPanelDescription(name='PANIC',model=''):
        """
        :param name: Name for Panel
        :param model: Model for the Panel
        :return: Panel Descritption Object
        """
        return PanelDescription(name,'vacca.alarms.VaccaPanic',
            model,sharedDataWrite={'HighlightInstruments':'devicesSelected'})

    @staticmethod
    def getDefaultIcon():
        """
        :return: The Default Icon Path.
        """
        path = fandango.objects.findModule('panic')+'/gui/icon/panic-6.png'
        print('VaccaPanic.getDefaultIcon():%s'%path)
        #path = vpath('image/icons/panic.gif')
        return path

    @classmethod
    def __test__(klass,arg=None):
        from vacca.utils import Qt
        qapp = Qt.QApplication([])
        i = klass()
        print i
        print klass
        i.show()

from doc import get_autodoc
__doc__ = get_autodoc(__name__,vars())
        
if __name__ == '__main__':
    import sys
    VaccaPanic.__test__(*sys.argv[1:])
        



#!/usr/bin/env python

#############################################################################
##
## This file is part of Taurus, a Tango User Interface Library
## 
## http://www.tango-controls.org/static/taurus/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
## 
## Taurus is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
## 
## Taurus is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
## 
## You should have received a copy of the GNU Lesser General Public License
## along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
###########################################################################

import fandango
from taurus.qt.qtgui.table import TaurusGrid
from taurus.qt.qtgui.taurusgui.utils import PanelDescription


class VaccaGrid(TaurusGrid):
    """ TaurusGrid enhanced to provide default PanelDescription """

    @staticmethod
    def getGridPanelDescription(grid):
        class_name = get_grid(grid)
        gridPanel = PanelDescription('VaccaGrid',
                                classname = 'vacca.VaccaGrid',
                                model = grid,
                                )
        return gridPanel

class VaccaVerticalGrid(TaurusGrid):

    @staticmethod
    def getVerticalGridPanelDescription(grid):
        class_name = get_grid(get_vertical_grid(grid))
        gridPanel = PanelDescription('VaccaVGrid',
                                classname = 'vacca.VaccaVerticalGrid',
                                model = get_vertical_grid(grid),
                                )
        return gridPanel

def get_empty_grid():
    tg = TaurusGrid()
    tg.showRowFrame(False)
    tg.showColumnFrame(False)
    tg.showAttributeLabels(False)
    tg.showAttributeUnits(False)
    return tg

def get_grid(grid):
    print 'get_grid(%s)'%str(grid.keys())
    tg = get_empty_grid()
    tg.setRowLabels(grid['row_labels'])
    tg.setColumnLabels(grid['column_labels'])
    tg.setModel(grid['model'])#,delayed=False)
    return tg

def get_vertical_grid(grid):
    vgrid = dict(grid)
    vgrid['column_labels'] = grid['row_labels']#'Gauges:VGCT|CCG|V-PEN,Pumps:IPCT|V-VARIP'
    vgrid['row_labels'] = grid['column_labels']
    vgrid['model'] = [s.replace('|State', '') for s in fandango.toSequence(
        grid['model'])]
    return vgrid #get_grid(vgrid)

from .doc import get_autodoc
__doc__ = get_autodoc(__name__,vars())


#!/usr/bin/env python

#############################################################################
##
## This file is part of Taurus, a Tango User Interface Library
## 
## http://www.tango-controls.org/static/taurus/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
## 
## Taurus is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
## 
## Taurus is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
## 
## You should have received a copy of the GNU Lesser General Public License
## along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
###########################################################################

"""
default.py.ini is a configuration file example of how to specify system-wide defaults for VaccaGUI 

After copying it to default.py this configuration file determines the default, permanent, pre-defined
options for the GUI; modify them to match your default gui options.

Every CONFIG file loaded by vaccagui will override those options redefined on it; 
use default.py for your common options and whatever.py for your in-place customization.
"""

import os,fandango,imp,vacca.utils
from taurus.qt.qtgui.button import TaurusLauncherButton
from vacca.utils import *
from fandango import matchCl,searchCl,replaceCl,CaselessDict,CaselessList,\
    get_matching_devices,get_matching_attributes,get_all_devices,check_device

LIGHTWEIGHT = True #This option will control whether heavy panels are loaded at startup or not

# Panic and PyTangoArchiving widgets will be added at the end of this file
try: 
  import panic
except: 
  panic = None
try:
  import PyTangoArchiving
except:
  PyTangoArchiving = None

###############################################################################
print('>'*20+' Loading vacca/default.py')

#ALL these variables can be re-defined in CONFIG FILE
GUI_NAME = 'VACCA'
WDIR = wdir() #imp.find_module('vacca')[1]+'/'
URL_HELP = 'http://computing.cells.es/services/controls/vacuum'
URL_LOGBOOK = 'http://logbook.cells.es/'
VACCA_LOGO = vpath('image/icons/AlbaLogo.png')
ORGANIZATION = 'TANGO'
ORGANIZATION_LOGO = vpath('image/icons/TangoLogo.png')

###############################################################################
# Synoptic file

JDRAW_FILE = '' #WDIR+'/examples/elinac/linac.jdw' #WDIR+'%s/%s.jdw'%(TARGET,TARGET)
#Enables/disables loading jdraw objects into tree
JDRAW_TREE = True 
#A method that does transformation on signals sent to other widgets.
JDRAW_HOOK = None #lambda s: apply_transform(get_coordinates(s)) 
# Enable multiple selection of objects in Synoptic
#from taurus.qt.qtgui.graphic import TaurusGraphicsScene
#TaurusGraphicsScene.ANY_ATTRIBUTE_SELECTS_DEVICE = True

###############################################################################
# Setup of the tree

#Domain/Target used to generate grids/trees ; domain may be regexp, target should not
DOMAIN = '*test*'
TARGET = DOMAIN
USE_DEVICE_TREE = True

#Devices not in JDraw or regular expression to be added to the tree
DEVICES = [d for d in get_all_devices() if not matchCl('^(tango|dserver)/*',d)]#map(bool,set(['%s/VC/ALL'%TARGET,'%s/CT/ALARMS'%TARGET,DEVICE,COMPOSER]))
EXTRA_DEVICES = []

#Custom tree branches are built using nested dictionaries and regular expressions (if empty a jive-like tree is built).
CUSTOM_TREE = {} 

# {'CT':'BL*(CT|ALARMS|PLC)$',
#  'FE':'FE*/VC/*',
#  'Valves': {'.OH':'*OH/PNV*',
#             'EH01':'*EH01/PNV*',},
#  'BAKEOUTS':'BL*(BAKE|BK)*',}

###############################################################################
# Device Panel setup

#PyStateComposer to get Vacuum Profile curves
COMPOSER = '' #'%s/vc/all'%DOMAIN
#Default device to appear in the DevicePanel
DEVICE = '' # 'sys/tg_test/1' ; tg_test caused segfault in some cases!?
USE_DEVICE_PANEL = True
PANEL_COMMAND = 'taurusdevicepanel --config-file='+wdir()+'default.py'

#Examples of Attribute filters to be applied to DevicePanel
AttributeFilters = {'V-PEN': ['pressure','channelstatus','controller'],}
AttributeFilters['EPS-TEST']=[ #You can distribute attributes in different tabs using tuples
    ('Status',['_READY','OPEN_','CLOSE_']),
    ('Signals',['.*_PT.*','was_','paas_','*RGA*']),
    ]
CommandFilters = {'V-PEN': (('on',()),('off',()),('setMode',('START','PROTECT'))),} #Second argument of tuple is the list of default arguments
IconMap = {'v-pen':WDIR+'image/equips/icon-pen.gif'} #Will be used in both panel and tree

## Optional:
## If you put filters in a separate file you can load a taurusdevicepanel.py with --config-file=filters.py option
## Then replace previous lines by:
# from vacca.filters import * 

###############################################################################
# Plot setup

#Pressure values to be added to the logaritmic trend
GAUGES = [] #['bl13/vc/vgct-01/p1','bl13/vc/vgct-01/p2'] 
#sorted(fandango.get_matching_attributes('%s/*/vgct*/p[12]'%DOMAIN.replace('BL','(BL|FE|ID)')))

###############################################################################
# Grid setup

#Grid showing all pressures in rows/columns
GRID = {
    'column_labels': '',
    'delayed': False,
    'frames': False,
    'model': [], 
    #'%s/VC/(IPCT|VGCT|CCGX)*/(P[12]|Pressure|State)$'%DOMAIN,'%s/V-[^/]*/[0-9]*/(Pressure|State)'%DOMAIN],
    'row_labels':'VcGauges(mbar):(VGCT|PEN), IonPumps(mbar):(IPCT|VARIP)',
    }

###############################################################################
# Extra widgets to appear in the NewPanel dialog

EXTRA_WIDGETS = []
# Definition for Panic/Archiving is done at the end of this file

###############################################################################
# Extra panels to be loaded at startup
# Extra apps to be added to the right-side Toolbar (loaded only on demand)

EXTRA_PANELS = {}
# Definition for Panic/Archiving is done at the end of this file

if not LIGHTWEIGHT:
  EXTRA_PANELS['Properties'] = vacca.VaccaPropTable.getPanelDescription('Properties',DEVICE) 


###############################################################################
# Define custom applets to be shown in the applets bar (the wide bar that
# contains the logos). To define an applet, instantiate an AppletDescription
# object (see documentation for the gblgui_utils module)
#

EXTRA_APPS = fandango.dicts.SortedDict()
from taurus.qt.qtgui.taurusgui.utils import PanelDescription, AppletDescription

EXTRA_APPS['Properties'] = {'class' : vacca.VaccaPropTable}

## Just another examples:

#EXTRA_APPS['Mambo'] = {'name': 'Mambo',
                #'class': lambda:os.system('mambo&'),
                #'icon': wdir('image/icons/Mambo-icon.png')}              
#EXTRA_APPS['DevicePanel'] = {'class' : vacca.VaccaPanel}
#EXTRA_APPS['Panic']= {'class' : vacca.VaccaPanic       }
#EXTRA_APPS['ExtraDock']= {'class' : Qt.QMainWindow       

#Each Applet can be described with a dictionary like this:
#   (name, classname=None, modulename=None, widgetname=None, 
#       sharedDataWrite=None, sharedDataRead=None, model=None, floating=True, **kwargs)
#For ExternalApp/VaccaActions use:
# {'$VarName':{'name':'$AppName','classname':'VaccaAction','model':['$Test','$/path/to/icon.png','$launcher']}}
    #'xrga':{'name':'RGA','classname':'VaccaAction','model':['RGA',WDIR+'image/equips/icon-rga.gif']+['rdesktop -g 1440x880 ctrga01']}
    
#from vacca.panel import VaccaAction
#xmambo = AppletDescription('Mambo',classname = 'vacca.panel.VaccaAction',
# model=["Archiving",wdir('vacca/image/widgets/ProfilePlot.png'),'mambo'],)
#xalarms = AppletDescription('Alarms',classname='vacca.panel.VaccaNewPanel',
# model=['Alarms',wdir('vacca/image/icons/panic.gif'),'panic.gui.AlarmGUI'])
#xsnap = AppletDescription('xSnap',classname='vacca.panel.VaccaAction', model=['Snap',vpath('image/widgets/VerticalGrid.jpg'),'ctsnaps'])
#xfinder = AppletDescription('xFinder',classname='vacca.panel.VaccaAction', model=['Attribute Finder',':actions/system-search.svg','ctfinder'])

#button =  TaurusLauncherButton(widget =
#button.setModel('a/b/c') #a device name, which will be given to the
# TaurusAttrForm when clicking

###############################################################################
# Define which External Applications are to be inserted.
# To define an external application, instantiate an ExternalApp object
# See TaurusMainWindow.addExternalAppLauncher for valid values of ExternalApp
#

TOOLBARS = [] #[(name,modulename.classname)]

MENUS = []
"""
MENUS = [('Tools',[('Jive',lambda:os.system('jive &'),None),])]
"""

from taurus.qt.qtgui.taurusgui.utils import ExternalApp, ToolBarDescription

#xvacca = ExternalApp(cmdargs=['konqueror',URL_HELP], text="Alba VACuum Controls Application", icon=WDIR+'/image/icons/cow-tux.png')
#logbook = ExternalApp(cmdargs=['konqueror %s'%URL_LOGBOOK], text="Logbook", icon=WDIR+"/image/icons/elog.png")
#xtrend = ExternalApp(cmdargs=['taurustrend','-a'], text="Trend")
#xjive = ExternalApp(cmdargs=['jive'], text="Jive")#, icon=WDIR+'/image/icons/cow-tux.png')
#xastor = ExternalApp(cmdargs=['astor'], text="Astor")#, icon=WDIR+'/image/icons/cow-tux.png')

EXTRA_TOOLS = [
  ('Jive',['jive'],vpath('image/icons/TangoLogo.png')),
  ('Trends',['taurustrend','-a'],vpath('image/widgets/PressureTrend.jpg')),
  ]

###############################################################################
# Add widgets/apps/panels for Alarms and Archiving

if panic:
  try:
    EXTRA_APPS['PANIC'] = {'name': 'PANIC',
        'class': vacca.VaccaPanic}

    EXTRA_WIDGETS.append(('panic.gui.AlarmGUI',wdir('image/icons/panic.gif')))
  
    if not LIGHTWEIGHT:
      EXTRA_PANELS['PANIC'] = PanelDescription(
        'PANIC','panic.gui.AlarmGUI',model='',#---
        sharedDataWrite={'HighlightInstruments':'devicesSelected'})
    
  except:
    print('Warning: Panic widgets not available')

if PyTangoArchiving: 
  try:
    EXTRA_WIDGETS.append(('PyTangoArchiving.widget.ArchivingBrowser.ArchivingBrowser',wdir('image/widgets/Archiving.png')))
    EXTRA_APPS['Finder'] = {'name': 'Finder','class': c,'icon': wdir('image/icons/search.png')}   

    if not LIGHTWEIGHT:
      try:
        import PyTangoArchiving.widget.browser
        c = PyTangoArchiving.widget.browser.ArchivingBrowser
      except:
        import PyTangoArchiving.widget.ArchivingBrowser
        c = PyTangoArchiving.widget.ArchivingBrowser.ModelSearchWidget

      EXTRA_PANELS['Finder'] = PanelDescription('Finder',w)
      
  except:
    print('Warning: PyTangoArchiving widgets not available')

__doc__ = vacca.get_autodoc(__name__,vars())



#!/usr/bin/env python

#############################################################################
##
## This file is part of Taurus, a Tango User Interface Library
## 
## http://www.tango-controls.org/static/taurus/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
## 
## Taurus is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
## 
## Taurus is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
## 
## You should have received a copy of the GNU Lesser General Public License
## along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

__all__ = ["TaurusDevTree", "TaurusSearchTree", "TaurusDevTreeOptions",
           "VaccaTree"]  # ,"SearchEdit"] #"TaurusTreeNode"]

import time, traceback
import fandango

try:
    import icons_dev_tree
except ImportError:
    icons_dev_tree = None

from taurus.external.qt import Qt

import taurus.core
from taurus.core.util.colors import DEVICE_STATE_PALETTE

from taurus.core.tango.search import get_alias_for_device
from taurus.qt.qtgui.base import TaurusBaseComponent, TaurusBaseWidget

TREE_ITEM_MIME_TYPE = 'application/x-qabstractitemmodeldatalist'
import taurus.qt.qtgui.tree.taurusdevicetree
from taurus.qt.qtgui.tree.taurusdevicetree import *
from taurus.qt.qtcore.mimetypes import *

if not hasattr(TaurusDevTree,'__pyqtSignals__'):
  TaurusDevTree.__pyqtSignals__ = (
   'modelChanged(const QString &)',
   'deviceSelected(QString)',
   'addAttrSelected(QStringList)',
   'removeAttrSelected(QStringList)',
   'refreshTree',
   'nodeFound')

# TaurusDevTree.setDefaultPanelClass is setting the widget to open when ContextMenu->OpenPanel is selected
TaurusSearchTree.setDefaultPanelClass = staticmethod(lambda c: TaurusDevTree.setDefaultPanelClass(c))
# TaurusDevTree.setDefaultPanelClass is setting the attributes to show when ContextMenu->ShowAttributes is selected
TaurusSearchTree.setDefaultAttrFilter = staticmethod(lambda c: TaurusDevTree.setDefaultAttrFilter(c))


class VaccaDevTree(taurus.qt.qtgui.tree.taurusdevicetree.TaurusDevTree, TaurusBaseWidget):
    ## @TODO: SAME CHANGE MUST BE ADDED TO TAURUSJDRAWGRAPHICS SELECTGRAPHICITEM!!!

    MAX_NODES_TO_UPDATE = 120
    
    def trace(self, msg):
        if self.TRACE_ALL or self.getLogLevel() in ('DEBUG', 40,):
            print 'TaurusDevTree.%s: %s' % (
            self.getLogLevel(), msg[:80] + '...')  # @TODO: use the taurus logger instead! ~~cpascual 20121121

    def findInTree(self, *a, **k):
        try:
            # Added to avoid recursion
            print('VaccaDevTree.findInTree(%s,%s)' % (a, k))
            if self.currentItem() and str(
                    self.getNodeDraggable()).lower().strip() == \
                    str(k.get('regexp', a[0])).lower().strip():
                return
            TaurusDevTree.findInTree(self, *a, **k)
        except:
            print traceback.format_exc()

    def getModelMimeData(self):
        """ Returns a MimeData object containing the model data. The default implementation
        fills the `TAURUS_MODEL_MIME_TYPE`. If the widget's Model class is
        Attribute or Device, it also fills `TAURUS_ATTR_MIME_TYPE` or
        `TAURUS_DEV_MIME_TYPE`, respectively
        """
        mimeData = Qt.QMimeData()
        node = self.currentItem()
        draggable = self.getNodeDraggable(node)
        if draggable:
            slashes = draggable.count('/') - draggable.count(':')
            # Set 'Text' To be compatible with other Widgets 'Text/plan':
            mimeData.setText(draggable)
            if slashes == 3:
                mimeData.setData(TAURUS_ATTR_MIME_TYPE, draggable)
            elif slashes == 2:
                mimeData.setData(TAURUS_DEV_MIME_TYPE, draggable)
                print 'mimeData is %s,%s' % (TAURUS_DEV_MIME_TYPE, draggable)
            elif slashes:
                mimeData.setData(TAURUS_MODEL_MIME_TYPE, draggable)
        print mimeData
        print map(str, mimeData.formats())
        return mimeData

    def setNodeTree(self, parent, diction, alias=False):
        """ OVERRIDEN TO ADD try/except CATCHES
        It has parent as argument to allow itself to be recursive
        Initializes the node tree from a dictionary {'Node0.0':{'Node1.0':None,'Node1.1':None}}
        """
        self.debug('In setNodeTree(%d,alias=%s) ...' % (len(diction), alias))
        if not hasattr(diction, 'keys'): diction = dict.fromkeys(diction)
        for node in sorted(diction.keys()):
            assert int(self.index) < 10000000000, 'TooManyIterations!'
            try:
                self.index = self.index + 1
                dev_alias = alias and str(node).count('/') == 2 and get_alias_for_device(node)
                text = '%s (%s)' % (node, dev_alias) if dev_alias else node
                if diction[node] and any(diction[node]):
                    item = self.createItem(parent, node, text)
                    self.setNodeTree(item, diction[node], alias)
                else:
                    item = self.createItem(parent, node, text)
            except:
                self.warning('setNodeTree(%s,%s) failed!: %s' % (parent, node, traceback.format_exc()))

    def getAllNodes(self):
        """ Returns a list with all node objects. """
        return self.item_index
      
    ###########################################################################
    # Update node colors
      
    def setIcons(self,dct={},root_name=None,regexps=True):
        '''
        This method change the icons depending of the status of the devices
        Dict is a dictionary with name of device and colors such as 
        {name_device:color,name_device2:color2}
        An alternative may be an icon name!
        '''
        try:
          state2color = lambda state: Qt.QColor(DEVICE_STATE_PALETTE.number(state))
          self.node_colors = getattr(self,'node_colors',{})
          dct = dict((k,v) for k,v in dct.items() if v!=self.node_colors.get(k,None))

          def update_node(node,key,dct):
              if hasattr(node,'CustomForeground'):
                  node.setForeground(0,Qt.QBrush(Qt.QColor(node.CustomForeground)))
              if hasattr(node,'CustomBackground'):
                  node.setBackground(0,Qt.QBrush(Qt.QColor(node.CustomBackground)))            
              elif hasattr(node,'StateBackground'):
                  node.setBackground(0,Qt.QBrush(state2color(dct[key])))
              if hasattr(node,'CustomIcon'):
                  node.setIcon(0,Qt.QIcon(node.CustomIcon))
              else:
                  if key.count('/')==2:
                      self.setStateIcon(node,dct and dct[key] or '')
              return

          nodes = self.getAllNodes()
          
          for name,node in nodes.iteritems():
              name = str(name).split()[0]
              if node.isHidden(): continue
              if regexps:
                  matches = [v for k,v in dct.items() if re.match(k.lower(),name.lower())]
                  if matches: 
                      update_node(node,name,{name:matches[0]})
              elif name in dct:
                  update_node(node,name,dct or {name:''})
                  
          self.node_colors.update(dct)
        except:
          self.warning('setIcons(): \n%s'%traceback.format_exc())
    
    def setStateIcon(self, child, color):
        color_codes = {
            '#00ff00,ON,OPEN,EXTRACT': ':/ICON_GREEN',
            "#ff0000,OFF,FAULT": ":/ICON_RED",
            "#ff8c00,ALARM": ":/ICON_ORANGE",
            "#ffffff,CLOSE,INSERT": ":/ICON_WHITE",
            "#80a0ff,MOVING,RUNNING": ":/ICON_BLUE",
            "#ffff00,STANDBY": ":/ICON_YELLOW",
            "#cccc7a,INIT": ":/ICON_BRAWN",
            "#ff00ff,DISABLE": ":/ICON_PINK",
            "#808080f,None,UNKNOWN": ":/ICON_GREY",
        }
        if icons_dev_tree is None:
            self.debug('In setStateIcon(%s,%s): Icons for states not available!'%(child,color))
            self.setStateBackground(child,color)
        else:
            icon = ":/ICON_WHITE"
            for states, code in color_codes.items():
                if str(color).upper() in states.upper():
                    icon = code
            self.debug('setStateIcon(%s) => %s' % (color, icon))
            icon = Qt.QIcon(icon)
            child.setIcon(0, icon)

    def setStateBackground(self, child, color):
        if not isinstance(color, Qt.QColor):
            if DEVICE_STATE_PALETTE.has(color):
                qc = Qt.QColor(*DEVICE_STATE_PALETTE.rgb(color))
            else:
                qc = Qt.QColor(color) if not fandango.isSequence(color) else Qt.QColor(*color)
        child.setBackground(0, Qt.QBrush(qc))
        
    def showNodeContextMenu(self,node,event):
        """
        A pop up menu will be shown with the available options. 
        Menus are managed using two tuple lists for each node: node.ContextMenu and node.ExpertMenu
        """
        obj = self.getNodeDraggable(node)
        position = event.globalPos()
        self.debug('showNodeContextMenu(%s)'%obj)
        if self.itemAt(position) is self.headerItem():
            node = self.headerItem()
            #node.ContextMenu = ['Search ...']
        if node is None:
            node = self
        else:
            if not hasattr(node,'ContextMenu'):
                node.ContextMenu=[]
            if not 'Search ...' in [k for k,a in node.ContextMenu]: ##Creating default menu
              
                if not hasattr(node,'ExpertMenu'): 
                  setattr(node,'ExpertMenu',self.ExpertMenu)
                  
                def addOption(menu,name,action):
                  if name not in [t[0] for t in menu]:
                    menu.append((name,action))
              
                # DEVICE NODE CONTEXT MENU
                if obj.count('/')==2:
                    
                    addOption(node.ContextMenu,"Open Panel", self.showPanel)
                    addOption(node.ContextMenu,"Show Attributes",self.addAttrToNode)
                    
                    if self.getNodeAdmin(node):
                        addOption(node.ContextMenu,"Go to %s"%self.getNodeAdmin(node),
                            (lambda p=self.getNodeAdmin(node): p and self.findInTree(p)))
                    
                    addOption(node.ContextMenu,'', None)

                    addOption(node.ContextMenu,"Show Properties", self.showProperties)
                    addOption(node.ContextMenu,"Test Device", self.test_device)
                    
                    try:
                      self.astor = fandango.Astor()
                      addOption(node.ContextMenu,'Start Server', self.start_server)
                      addOption(node.ContextMenu,'Stop Server', self.stop_server)
                      addOption(node.ContextMenu,'Device Info', self.device_info)
                    except:
                      self.warning('fandango.Astor() not available to start/stop devices')
                        
                    node.ContextMenu.append(('',None))
                    addOption(node.ExpertMenu,"Show ALL Attributes", lambda s=self:s.addAttrToNode(full=True))
                    
                # ATTRIBUTE NODE CONTEXT MENU
                elif obj.count('/')==3:
                    for k,v in self.AttributeMenu:
                        self.debug('Adding action %s'%k)
                        if type(v) is str and hasattr(self,v):
                            node.ContextMenu.append((k, getattr(self,v)))
                        else:
                            node.ContextMenu.append((k, lambda s=self.getNodeAlias(node): v(s)))
                    #node.ContextMenu.append(("add to Trends", self.addToPlot))
                    #node.ContextMenu.append(("remove from Trends", self.removeFromPlot))
                    node.ContextMenu.append(('',None))
                    
                #node.ContextMenu.append(("Expand Node", self.expandNode))
                #node.ContextMenu.append(("Collapse Node", self.collapseNode))
                
                if node.isExpanded() and node.childCount()<10 and all(self.getNodeText(node.child(j)).count('/')==2 for j in range(node.childCount())):
                    node.ContextMenu.append(("Show Attributes", lambda n=node,s=self: [s.addAttrToNode(n.child(j)) for j in range(n.childCount())]))
                node.ContextMenu.append(("Search ...",\
                    lambda: self.findInTree(str(Qt.QInputDialog.getText(self,'Search ...','Write a part of the name',Qt.QLineEdit.Normal)[0]))
                    ))
        #configDialogAction = menu.addAction("Refresh Tree")
        #self.connect(configDialogAction, Qt.SIGNAL("triggered()"), self.refreshTree)
        menu = Qt.QMenu(self)
        
        if hasattr(node,'ContextMenu'):
            last_was_separator = True
            for t in (type(node.ContextMenu) is dict and node.ContextMenu.items() or node.ContextMenu):
                try:
                    k,action = t
                    if k:
                        configDialogAction = menu.addAction(k)
                        if action: self.connect(configDialogAction, Qt.SIGNAL("triggered()"), action)
                        else: configDialogAction.setEnabled(False)
                        last_was_separator = False
                    elif not last_was_separator: 
                        menu.addSeparator()
                        last_was_separator = True
                except Exception,e: 
                    self.warning('Unable to add Menu Action: %s:%s'%(t,e))
        
        if hasattr(node,'ExpertMenu'):
            menu.addSeparator()
            expert = menu.addMenu('Expert')
            #expert.addSeparator()
            last_was_separator = True
            for t in (type(node.ContextMenu) is dict and node.ExpertMenu.items() or node.ExpertMenu):
                try:
                    k,action = t
                    if k:
                        configDialogAction = expert.addAction(k)
                        if action: self.connect(configDialogAction, Qt.SIGNAL("triggered()"), action)
                        else: configDialogAction.setEnabled(False)
                        last_was_separator = False
                    elif not last_was_separator: 
                        expert.addSeparator()
                        last_was_separator = True
                except Exception,e: 
                    self.warning('Unable to add Expert Action: %s:%s'%(t,e))            
        #menu.addSeparator()
        menu.exec_(event.globalPos())
        del menu
        
    def start_server(self, device=None):
        """
        Allow start Servers.
        :param device: DeviceName
        :return:
        """
        device = device or self.getNodeDeviceName()
        self.astor.load_by_name(device)
        ss = self.astor.get_device_server(device)
        text, ok = Qt.QInputDialog.getText(self, 'Start Server', 'Start %s at '
                                                                 'host ...'
                                           % ss, Qt.QLineEdit.Normal,
                                           self.astor[ss].host)
        if ok:
            return self.astor.start_servers(ss, host=str(text))
        else:
            return False

    def stop_server(self, device=None):
        """
        Allow stop Servers
        :param device: DeviceName
        :return:
        """
        device = device or self.getNodeDeviceName()
        self.astor.load_by_name(device)
        ss = self.astor.get_device_server(device)
        v = Qt.QMessageBox.warning(self, 'Stop Server', 
            '%s will be killed!, Are you sure?'%ss,
            Qt.QMessageBox.Yes | Qt.QMessageBox.No)
        if v == Qt.QMessageBox.Yes:
            return self.astor.stop_servers(ss)
        else:
            return False

    def device_info(self, device=None):
        """
        Show a the Device Info
        :param device: DeviceName
        :return:
        """
        device = device or self.getNodeDeviceName()
        di = fandango.tango.get_device_info(device)
        txt = '\n'.join('%s : %s' % (k, getattr(di, k)) for k in
                        'name dev_class server host level exported started stopped PID'.split())
        v = Qt.QMessageBox.information(self, device, txt, Qt.QMessageBox.Ok)
        
    def test_device(self):
        import os
        device = str(self.getNodeDeviceName())
        if not device: return
        comm = 'tg_devtest %s &'%device
        os.system(comm)        

class VaccaTree(TaurusSearchTree):
    """
    It is a class that inherits from TaurusSearchTree.
    Allow show the devices and start/stop it with the right button (
    expandable menu)

    """

    # This slots are overloaded here because they are not yet in the last taurus package. Once it will be included in TaurusSearchTree than it can be removed.
    # The slots are needed because the method_forwarder method is not seen from the SharedDataManager side.

    # def __getattr__(self,attr):
    # if attr!='tree':
    # return getattr(self.tree,attr)

    ## DO  NOT REMOVE YET!!
    def setTangoHost(self, *a, **k):
        self.tree.setTangoHost(*a, **k)

    def addModels(self, *a, **k):
        self.tree.addModels(*a, **k)

    def setModel(self, *a, **k):
        self.tree.setModel(*a, **k)

    def setModelCheck(self, *a, **k):
        self.tree.setModelCheck(*a, **k)

    def setTree(self, *a, **k):
        self.tree.setTree(*a, **k)

    def expandAll(self, *a, **k):
        self.tree.expandAll(*a, **k)

    def loadTree(self, *a, **k):
        print('VaccaTree.loadTree(...)')
        self.tree.loadTree(*a, **k)

    @staticmethod
    def setDefaultPanelClass(*a, **k):
        TaurusDevTree.setDefaultPanelClass(*a, **k)

    @staticmethod
    def setDefaultAttrFilter(*a, **k):
        TaurusDevTree.setDefaultAttrFilter(*a, **k)

    @classmethod
    def setIconMap(klass, iconMap):
        TaurusDevTree.setIconMap(iconMap)

    def defineStyle(self):
        # print('In TaurusSearchTree.defineStyle()')
        self.setWindowTitle('VaccaTree')
        self.setLayout(Qt.QVBoxLayout())
        self.edit = TaurusDevTreeOptions(self)
        self.tree = VaccaDevTree(self)
        self.setLogLevel(self.Warning)
        self.tree.setLogLevel(self.Warning)

        self.layout().addWidget(self.edit)
        self.layout().addWidget(self.tree)
        self.registerConfigDelegate(self.tree)
        # Event forwarding ...
        for signal in TaurusDevTree.__pyqtSignals__:        
            Qt.QObject.connect(self.tree,
                               Qt.SIGNAL(signal),
                               lambda args,
                                      f=self,
                                      s=signal: f.emit(Qt.SIGNAL(s), args))
            
        self.edit.connectWithTree(self.tree)
        self.statetimer = Qt.QTimer(self)
        self.connect(self.statetimer, Qt.SIGNAL('timeout()'), self.updateStates)
        self.statetimer.start(333)
        return

    def updateStates(self):
        try:
            self.debug('On VaccaTree.updateStates()')
            if not hasattr(self,'_statecount'): self._statecount = 0
            if not self._statecount: 
                self._nodes2update = self.tree.getAllNodes().keys()
                self._allexported = fandango.get_all_devices(exported=True)

            if len(self._nodes2update) > VaccaDevTree.MAX_NODES_TO_UPDATE:
              if getattr(self,'_coloured',True):
                self._coloured = False
                whites = dict((k,'CLOSE') for k in self._nodes2update)
                self.tree.setIcons(whites,regexps=False)
                self.warning('too many tree nodes (%s) to update!'%(len(self._nodes2update)))
              else: pass #Not trivial
            else:
              t0,dct = time.time(),{}
              while time.time() < t0+2e-3:
                  if self._statecount>=len(self._nodes2update):
                      self._statecount = 0
                  k = self._nodes2update[self._statecount]
                  self._statecount+=1 #< must be here
                  try:
                      if k.count('/') == k.count(':')+2:
                          x = k in self._allexported
                          self.debug('On VaccaTree.updateStates(%s,%s,%s)'%(self._statecount,k,x))
                          if not x:
                            self.debug('\t%s not exported!'%k)
                            dct[k] = 'UNKNOWN'
                          else:
                            dct[k] = str(taurus.Attribute(k+'/State').read().value)
                  except:
                      self.info('updateStates(%s) failed!: %s'%(k,traceback.format_exc()))
              if dct:
                  self.tree.setIcons(dct,regexps=False)

        except:
            self.warning('On VaccaTree.updateStates(): %s' % traceback.format_exc())
            
    @staticmethod
    def getDefaultIcon():
        """
        :return: The Default Icon Path.
        """
        path = 'image/widgets/Tree.png'
        return path

from vacca.doc import get_autodoc
__doc__ = get_autodoc(__name__, vars())

###############################################################################

def main(args=[]):
    import sys
    args = args or sys.argv[1:]
    qapp = Qt.QApplication(sys.argv)
    vt = VaccaTree()
    vt.setModel(args or fandango.get_all_devices())
    vt.show()
    qapp.exec_()

if __name__=='__main__':
    main()




#!/usr/bin/env python

#############################################################################
##
## This file is part of Taurus, a Tango User Interface Library
## 
## http://www.tango-controls.org/static/taurus/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
## 
## Taurus is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
## 
## Taurus is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
## 
## You should have received a copy of the GNU Lesser General Public License
## along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

__doc__ = """vacca.utils python module, by srubio@cells.es

vacca.utils contains several methods and classes for:
 
 - manage environment variables
 - loading config files
 - manipulate Qt color palettes
 - custom Taurus event filters
 
This file is part of Tango Control System

Environment Variables
^^^^^^^^^^^^^^^^^^^^^

Variables can be declared in OS or as Tango.VACCA free properties

VACCA_CONFIG: *get_config_file()*

    The config file to be loaded, it can be absolute path or 
    relative to VACCA_DIR.
    
    e.g. /homelocal/sicilia/applications/vacca/tbl2401.py
    
VACCA_DIR: *wdir('path/to/icon')* 

    The folder were VACCA should be executed, all paths within 
    the application will be relative to it. It includes all custom icons 
    that are not part of the vacca package.

    If VACCA_CONFIG is defined, it defaults to the path to config file.
    At ALBA installations it defaults to the path of vaccagui library.    
    Extra modules or plugins to be imported will be loaded from this directory.
    
    e.g. /homelocal/sicilia/applications/vacca/
    
VACCA_PATH: *vpath('path/to/icon')*

    Not mandatory, it defaults to the path of vacca library.
    Used for launchers and default icons of vacca widgets.

    e.g. /homelocal/sicilia/lib/python/site-packages/vacca

"""

import os,sys,traceback,imp,time,re
import fandango
from fandango import printf,get_database,first,\
    isString,isRegexp,clmatch
from fandango.dicts import SortedDict
from taurus.external.qt import Qt
from taurus.external.qt import Qwt5

try:
  from taurus.core import TaurusDatabase as TaurusAuthority
  print('Using Taurus <=3')
except:
  from taurus.core import TaurusAuthority
  print('Using Taurus >=4')

from taurus.qt.qtgui.base import TaurusBaseComponent
from taurus.core.util.eventfilters import ONLY_CHANGE_AND_PERIODIC
from taurus.qt.qtgui.taurusgui.utils import PanelDescription
from taurus.qt.qtgui.taurusgui import TaurusGui

###############################################################################

DB_HOST = fandango.tango.get_tango_host().split(':')[0]

def get_env_variable(var,default=''):
    """ It will try to get value from OS, then Tango.VACCA, then default """
    v = os.environ.get(var,None)
    if v: return v
    try:
      v = fandango.get_database().get_property('VACCA',var)[var]
      v =  v[0] if v and len(v) == 1 else v
      if v: return v
    except:
      traceback.print_exc()
    return default

def get_class_property(dev_class,prop,extract=False):
    db = fandango.get_database()
    props = db.get_class_property(dev_class,prop)
    if prop in props:
        v = props[prop]
        if fandango.isSequence(v): v = list(v)
        if v == ['']: return []
        if extract and fandango.isSequence(v) and len(v)==1:
            v = v[0]
        return v
    return []
   
def get_vacca_property(prop,extract=False):
    #print('get_vacca_property(%s)'%prop)
    db = fandango.get_database()
    props = db.get_property('VACCA',prop)
    if prop in props:
        v = props[prop]
        if fandango.isSequence(v): v = list(v)
        else: v = [v]
        if not any(v): return []
        for i,vv in enumerate(v):
          v[i] = vv = replace_env(vv,'VACCA_PATH')
          v[i] = vv = replace_env(vv,'VACCA_DIR')
        if extract and len(v)==1:
            v = v[0]
        return v
    return [] if not extract else ''

def replace_env(text,var=('VACCA_DIR','VACCA_PATH'),value=None):
    """
    Replaces each occurrence of $ENV_VARIABLE by its value.
    Var names can be a sequence.
    """
    if not fandango.isString(text):
      return text
    elif fandango.isSequence(var):
      return reduce(replace_env,[text]+list(var))
    else:
      if value is None: value = os.getenv(var)
      e = '[$][(]?'+var+'[)]?(?![a-zA-Z_0-9])'
      return re.sub(e,str(value),text)
  
###############################################################################

global VACCA_PATH
VACCA_PATH = get_env_variable('VACCA_PATH', imp.find_module('vacca')[1])
"""
VACCA_PATH: *vpath('path/to/icon')*

    Path to the VACCA library parent folder, tipically used for icons of vacca package
    Note, vpath(folder) will append /vacca to VACCA_PATH, to get internal vacca files.

    e.g. VACCA_PATH = /homelocal/sicilia/lib/python/site-packages
    e.g. vpath = /homelocal/sicilia/lib/python/site-packages/vacca
"""

os.environ['VACCA_PATH'] = VACCA_PATH
if VACCA_PATH+'/ini' not in sys.path:
  sys.path.append(VACCA_PATH+'/ini')

global VACCA_DIR
VACCA_DIR = get_env_variable('VACCA_DIR', get_env_variable('WORKING_DIR',''))
VACCA_DIR = replace_env(VACCA_DIR,'VACCA_PATH',VACCA_PATH) or VACCA_PATH
"""
VACCA_DIR: *wdir('path/to/icon')* 

    The folder were VACCA should be executed, all paths within 
    the application will be relative to it. It includes all custom icons 
    that are not part of the vacca package.
    
    Extra modules or plugins to be imported will be loaded from this directory.
    
    e.g. /homelocal/sicilia/applications/vacca/
"""

#Needed for backwards compatibility
#VACCA_DIR NOT FORCED HERE TO ALLOW OVERRIDE BY VACCA_CONFIG
WORKING_DIR = VACCA_DIR
DEFAULT_PATH =  WORKING_DIR

global VACCA_CONFIG
VACCA_CONFIG = get_env_variable('VACCA_CONFIG',DB_HOST+'.py')
VACCA_CONFIG = replace_env(replace_env(VACCA_CONFIG,'VACCA_DIR',VACCA_DIR),
                           'VACCA_PATH',VACCA_PATH)
VACCA_CONFIG = VACCA_CONFIG if os.path.isfile(VACCA_CONFIG) else VACCA_PATH+'/default.py'
"""
VACCA_CONFIG: 

    The config file to be loaded, it can be absolute path or 
    relative to WORKING_DIR. *get_config_file()* will import it and 
    return the content as a python module.
    
    e.g. /homelocal/sicilia/applications/vacca/tbl2401.py
"""
  
###############################################################################

def _joiner(a,s,b):
    return '%s%s%s'%(a or '',s if a and b and s not in (a[-1],b[0]) else '',b or '')

def wdir(s=''):
    #It tries to convert all paths to absolute
    global VACCA_DIR
    d = VACCA_DIR = os.getenv('VACCA_DIR')
    if not d and not os.path.exists(s): 
        d = os.path.dirname(VACCA_CONFIG)
    d = _joiner(d,'/',s)
    return d

def vpath(s=''):
    #It tries to convert all paths to absolute
    global VACCA_PATH
    VACCA_PATH = os.getenv('VACCA_PATH')
    s2 = _joiner(VACCA_PATH,'/',s)
    return s2
  
def expand_device_list(expressions):
    """
    Expands a list of regular expressions into a list of devices
    
    It adds 2 new reserverd characters:
     - ! , at the beginning of the expression is a negative lookup
     - ? , at the beginning of the expression, matches only exported devices
    """
    devices,ins,outs = [],set(),set()
    if isString(expressions):
      expressions = (','.join(expressions.split())).split(',')
    [(ins,outs)['!' in s].add(s.strip().lower()) 
            for s in expressions if s.strip()]

    for s in ins:
      try:
        if isRegexp(s):
          ds = fandango.get_matching_devices(s.strip('?'),
                exported='?' in s)
          devices.extend(d for d in ds 
                if not clmatch('*dserver/*/*',d))
        else:
          devices.append(s)

      except Exception,e:
        print(e,traceback.format_exc())
        
    #print('expand_device_list(%s): %s'%(
        #expressions,str(devices)[:80]))
    
    for o in outs:       
      try:
        n = len(devices)
        devices = [d for d in devices 
            if not fandango.matchCl(o.strip('!'),d)]
      except:
        print(e,traceback.format_exc())
        
    #print('expand_device_list(%s): %s'%(expressions,len(devices)))
    return sorted(devices)
  
###############################################################################
  
global VACCA_PROFILES
VACCA_PROFILES=[]
"""
VACCA_PROFILES: 

    Sets of environment variables declared as free properties
    in the Tango Database.
    
"""

def create_config_properties():
    global VACCA_PATH
    configs = ['DEFAULT']
    get_database().put_property('VACCA',{
        'VaccaConfigs':configs})
    get_database().put_property('VACCA',{
        'DEFAULT':['VACCA_CONFIG='+VACCA_PATH+'/default.py']})
    print('Created default VACCA properties')
    return get_config_properties()    

def load_config_properties(config,export=True):
    """
    It will load env variables from VACCA.$config property
    
    NOT USED!?
    """
    print('In load_config_properties(%s)'%config)
    props = get_vacca_property(config)
    props = [l.split('#')[0].strip() for l in props]
    props = dict(l.split('=',1) for l in props if l)
    if export:
        global VACCA_CONFIG,VACCA_DIR,VACCA_PATH

        if not VACCA_DIR and 'VACCA_DIR' in props:
            VACCA_DIR=props['VACCA_DIR']

        if not VACCA_PATH and 'VACCA_PATH' in props:
            VACCA_CONFIG=props['VACCA_PATH']            

        if not VACCA_CONFIG and 'VACCA_CONFIG' in props:
            
            l=props['VACCA_CONFIG']
            for p in ('VACCA_DIR','VACCA_PATH'):
                l=replace_env(l,p,globals()[p])
            VACCA_CONFIG = props['VACCA_CONFIG'] = l
            
        if not VACCA_DIR and '/' in VACCA_CONFIG:
            VACCA_DIR=os.path.dirname(VACCA_CONFIG)
    print(props)
    return props
  
def get_config_properties(config=''):
    """
    This method will get the list of VACCA free properties.
    If a config is passed, it will parse its contents as a dictionary.
    A new declaration of VACCA_CONFIG will be parsed as CONFIG_FILE instead.
    """
    #print('vacca.utils.get_config_properties(%s)'%config)
    global VACCA_PROFILES
    if not VACCA_PROFILES:
        VACCA_PROFILES = SortedDict()
        for k in get_vacca_property('VaccaConfigs',False):
          VACCA_PROFILES[k] = {}
    if not config:
        r = VACCA_PROFILES
    else:
        if not VACCA_PROFILES.get(config):
          VACCA_PROFILES[config] = [l.split('#')[0].strip() \
            for l in get_vacca_property(config,False)]
        config = VACCA_PROFILES[config]
        r = dict(l.split('=',1) for l in config if l)
        if 'VACCA_CONFIG' in r and 'CONFIG_FILE' not in r:
          r['CONFIG_FILE'] = r.pop('VACCA_CONFIG')
    return r

def get_config_file(config=None):
    """
    This method parses properties and returns the value of CONFIG_FILE.
    If undefined, VACCA_CONFIG is returned as fallback.
    """
    global VACCA_CONFIG,VACCA_DIR
    print('-')*80
    if not config:
        VACCA_CONFIG = get_env_variable('VACCA_CONFIG',DB_HOST)
        configs = get_config_properties()
        if VACCA_CONFIG in configs:
          CONFIG_FILE = get_config_properties(VACCA_CONFIG).get('CONFIG_FILE',VACCA_CONFIG)
        else:
          CONFIG_FILE = VACCA_CONFIG
        #CONFIG_FILE = CONFIG_FILE.replace('$VACCA_DIR',VACCA_DIR or '').replace('$VACCA_PATH',VACCA_PATH or '')
        CONFIG_FILE = replace_env(CONFIG_FILE,'VACCA_DIR',VACCA_DIR or '')
        CONFIG_FILE = replace_env(CONFIG_FILE,'VACCA_PATH',VACCA_PATH or '')
    else:
        CONFIG_FILE = config
    print('In get_config_file(%s)'%CONFIG_FILE)

    if not CONFIG_FILE.startswith('/'):
      if not os.path.isfile(CONFIG_FILE):
        cf = os.path.abspath(CONFIG_FILE)
        CONFIG_FILE = cf if os.path.isfile(cf) else wdir(CONFIG_FILE)
    print CONFIG_FILE
    
    if not fandango.linos.file_exists(CONFIG_FILE):
        if fandango.linos.file_exists(CONFIG_FILE+'.py'):
            CONFIG_FILE += '.py'
        else:
            print('WARNING: %s config file not found!'%CONFIG_FILE)
            return

    print 'VaccaGUI: Loading %s ...' % CONFIG_FILE
    try: 
        CONFIG = globals().get(CONFIG_FILE,fandango.objects.loadModule(CONFIG_FILE))
        #print '\n'.join(sorted('\t%s:\t%s'%(k,getattr(CONFIG,k)) for k in dir(CONFIG) if not k.startswith('_')))
    except: 
        print traceback.format_exc()
        raise Exception,'Unable to load %s' % CONFIG_FILE
    print('Out of get_config_file(...)\n')
    return CONFIG
  
###############################################################################
# Miscellaneous methods to be used with Taurus GUI
  
def get_os_launcher(cmd,args=[]):
    """
    Returns a Linux launcher method for "cmd".
    @TODO: Study Taurus.ExternalApp and use the same methods used there.
    """
    import os
    cmd += ' '.join(args)
    if '&' not in cmd: cmd+='&'
    f = (lambda : os.system(cmd))
    return f
  
def get_main_window(app=None):
    app = app or Qt.QApplication.instance()
    main = fandango.first((a for a in app.allWidgets() if isinstance(a,Qt.QMainWindow)),default=None)
    return main

def get_taurus_gui(app=None):
    try:
        #find actual Taurusgui Instance
        app = app or Qt.QApplication.instance()
        assert app
    except:
        traceback.print_exc()
        return
    
    try:
        taurusgui = None
        widgets = app.allWidgets()
        for widget in widgets:
            if isinstance(widget, TaurusGui):
                taurusgui = widget  
        return taurusgui
    except:
        traceback.print_exc()
        return

def get_shared_data_manager():
    """
    sdm = get_shared_data_manager() #Qt.qApp.SDM Singletone
    sdm.connectReader('channelName',objectInstance.method,readOnConnect=False) #Using instantiated object/function instead of names!!
    sdm.connectWriter('SelectedInstrument',out_hook,'modelChanged') #'Channel', object, 'qt signal name'
    """
    try:
        return Qt.qApp.SDM
    except:
        print('Shared Data Manager is not available! (no TaurusGUI instance?)')
        return None
    
def get_shared_data_signals():
    """
    Return the dictionary of signals and registered Qt objects.
    """
    try:
        sdm = get_shared_data_manager()
        signals = {}
        for k,v in sdm._SharedDataManager__models.items():
            signals[k] = {'readers':[],'writers':[]}
            for o,s in v._DataModel__readerSlots:
                signals[k]['readers'].append((o(),s))
            for o,s in v._DataModel__writerSignals:
                signals[k]['writers'].append((o(),s))
        return signals
    except:
        traceback.print_exc()
        return {}

                
###############################################################################
# Methods for managing palettes
    
def get_solid_brush(r, g, b, pattern=None):
    from taurus.external.qt import Qt
    brush = Qt.QBrush(Qt.QColor(r, g, b))
    brush.setStyle(Qt.Qt.SolidPattern if pattern is None else pattern)
    return brush

def get_White_palette(palette = None):
    from taurus.external.qt import Qt
    if palette is None:
        palette = Qt.QPalette()

    for e in (Qt.QPalette.Base,Qt.QPalette.Window): #QtGui.QPalette.BrightText,QtGui.QPalette.ButtonText,Qt.QPalette.Button
        for r in (Qt.QPalette.Active, Qt.QPalette.Inactive,
                  Qt.QPalette.Disabled):
            palette.setBrush(r, e, get_solid_brush(255, 255, 255))
    return palette
    
def get_halfWhite_palette(palette = None):
    from taurus.external.qt import Qt
    if palette is None:
        palette = Qt.QPalette()
    palette.setBrush(Qt.QPalette.Active, Qt.QPalette.Button,
                     get_solid_brush(255, 255, 255))
    palette.setBrush(Qt.QPalette.Inactive, Qt.QPalette.Button,
                     get_solid_brush(255, 255, 255))
    palette.setBrush(Qt.QPalette.Disabled, Qt.QPalette.Button,
                     get_solid_brush(255, 255, 255))
    return palette

def get_fullWhite_palette(palette = None):
    from taurus.external.qt import Qt
    if palette is None:
        palette = Qt.QPalette()
    palette = get_halfWhite_palette(palette)
    palette = get_White_palette(palette)
    return palette

###############################################################################
# Methods for using Taurus Event Filters

import fandango as f

class EventFilter(object):
    """ Generic event filter for taurus """
    def __init__(self,condition=None):
        self.condition = condition
    def __call__(self,s,t,v):
        if (self.condition() if f.isCallable(self.condition) else self.condition):
            return s,t,v
        else:
            return None

class EventCounter(EventFilter):
    """ Generic event counter for taurus """
    @f.Catched
    def __call__(self,s,t,v):
        try: src = s.getFullName()
        except: src = str(s)
        if not hasattr(self,'counts'): self.counts = f.defaultdict(int)
        self.counts[src] += 1
        return s,t,v
        
class OnChangeOrTimeEventFilter(EventFilter):
    """
    The instances of this class will be callables that can be used as filters
    of repeated-value events.
    If the event type is Change or Periodic, it will only pass when its
    evt_value.value is different from that of the last event received from the
    same source and type. If evt_value.value is not available in the current
    event, the whole evt_value is used for comparison and as a future reference.
    But!, if setMaxDeltaTime is set, then any event that occurred more than Delta Time seconds 
    after the last one will be allowed to pass through the chain.
    

    This is useful to avoid processing repetitive events.

    Note that you cannot use this class as a filter: you need to use an
    instance of it.

    Note 2: Use a different instance each time you insert this filter
    into a different widget unless you *really* know what you are doing.

    Example of usage::
        evf = OnChangeOrTimeEventFilter()
        evf.setDeltaTime(60.)
        filters = [evf, IGNORE_CONFIG]
        MyTaurusComponentInstance.setEventFilters(filters,preqt=True)
        filterEvent(s, t, v, filters)
    """
    def __init__(self):
        self._lastValues = {}
        self._lastTimes = {}
        self.deltaValue = 0
        self.deltaTime = 0
        
    def setDeltaValue(self,mindiff):
        self.deltaValue = mindiff
        
    def setDeltaTime(self,mindiff):
        self.deltaTime = mindiff
        
    @f.Catched
    def __call__(self, s, t, v):
        # restrict this  filter only to change and periodic events.
        if ONLY_CHANGE_AND_PERIODIC(s, t, v) is None: return s,t,v
        # block event if we recorded one before with same src, type and v.value
        new_value = getattr(v, 'value', v)
        try: new_time = ctime2time(v.time)
        except: new_time = time.time()
        try:
            if self._lastValues[(s, t)] == new_value and (not self.deltaTime or new_time<self._lastTimes[s]+self.deltaTime):
                return None
        except KeyError: pass
        # if it was't blocked, store src, type and v.value for future checks
        self._lastValues[(s, t)] =  new_value
        self._lastTimes[s] = new_time
        return s, t, v
    
    @staticmethod
    def __test__():
        import taurus,time
        from vacca.utils import EventFilter,EventCounter,OnChangeOrTimeEventFilter,MyTaurusComponent
        tbc = MyTaurusComponent('test')
        attr1,attr2 = 'test/sim/tonto/temperature','test/sim/tonto/temperate'
        tbc.setMyHandler(lambda *args:fandango.printf(str(
                    (time.ctime(),getattr(args[-1],'value',args[-1]),
                    taurus.Attribute(attr2).read().value))))
        evf = OnChangeOrTimeEventFilter()
        evf.setDeltaTime(10.)
        tbc.setEventFilters((evf,))
        tbc.setEventBufferPeriod(3.)
        tbc.setModel(attr1)
        ta = tbc.getModelObj()
        ta.changePollingPeriod(1000)
        
###############################################################################
# QT Helping Classes

from taurus.qt.qtcore.mimetypes import TAURUS_ATTR_MIME_TYPE, TAURUS_DEV_MIME_TYPE, TAURUS_MODEL_MIME_TYPE, TAURUS_MODEL_LIST_MIME_TYPE
      
def add_menu(menuname,actions):
    '''
    The arguments must be a list of tuples with action names and action methods.
    
    e.g.:
    EXTRA_MENUS = [
    ('Tools',[
        ('LTB Gui',lambda:os.system('vacca_LTB &'),None),
        ('Jive',lambda:os.system('jive &'),None),
        ('Astor',lambda:os.system('astor &'),None),
        ('EPS Gui',lambda:os.system('alba_EPS &'),None),
        ('Valves',lambda:valves.ValvesChooser().show(),None),
        ('TaurusTrend',lambda:os.system('taurustrend -a &'),None),
        ]),
    '''
    print 'Adding new menu %s'%menuname
    self = get_main_window()
    MenuBar = self.menuBar()
    newmenu = Qt.QMenu(MenuBar)
    newmenu.setTitle(menuname)
    newmenu.setObjectName(menuname)
    newaction = Qt.QAction(MenuBar)
    for name,method,mnemonic in actions:
        act = Qt.QAction(newmenu)
        act.setObjectName(name)
        act.setText(name)
        newmenu.addAction(act)
        Qt.QObject.connect(act,Qt.SIGNAL("triggered()"),method)            

    #self.MenuBar.addAction(newmenu.menuAction())
    MenuBar.insertMenu(newaction,newmenu)
    return

class WidgetAcceptDrops(object):
    """
    Template for Taurus widgets accepting drops, deprecated by fandango.qt.Dropable
    """
    TAURUS_DEV_MIME_TYPE = TAURUS_DEV_MIME_TYPE
    TAURUS_ATTR_MIME_TYPE = TAURUS_ATTR_MIME_TYPE
    TAURUS_MODEL_MIME_TYPE = TAURUS_MODEL_MIME_TYPE
    TAURUS_MODEL_LIST_MIME_TYPE = TAURUS_MODEL_LIST_MIME_TYPE
    TEXT_MIME_TYPE = 'text/plain'
    _supportedMimeTypes = None
    
    def checkDrops(self): 
        try: self.setAcceptDrops(True)
        except: pass
    def getSupportedMimeTypes(self):
        self.checkDrops() 
        if self._supportedMimeTypes is not None: return self._supportedMimeTypes
        else: return [TEXT]
    def setSupportedMimeTypes(self,mimes):
        self._supportedMimeTypes = mimes
    def setDropEventCallback(self,callback):
        self._dropeventcallback = callback
        
    # ENABLING DROP OF DEVICE NAMES :
    def checkSupportedMimeType(self,event,accept=False): 
        if any(t in self.getSupportedMimeTypes() for t in event.mimeData().formats()):
            if accept: event.acceptProposedAction()
            return True
        return False
    def dragEnterEvent(self,event): self.checkSupportedMimeType(event,accept=True)
    def dragMoveEvent(self,event): event.acceptProposedAction()
            
    def dropEvent(self, event):
        '''reimplemented to support drag&drop of models. See :class:`QWidget`'''
        print 'In dropEvent',event
        for t in self.getSupportedMimeTypes():
            try:
                t = str(event.mimeData().data(t))
                try:
                    if t.strip():
                        print t
                        self._dropeventcallback(t)
                        return
                except: self.info(traceback.warning_exc())
            except: pass
        self.debug('Invalid data (%s) for MIMETYPE'%(repr(d)))
        event.acceptProposedAction()
        
class Draggable(object):
    """
    Template for Taurus widgets accepting drops, deprecated by fandango.qt.Draggable
    """
    
    def __init__(self):
        self._drageventcallback = lambda s=self:s.text() if hasattr(s,'text') else ''
    
    def setDragEventCallback(self,hook):
        self._drageventcallback = hook
    
    def mousePressEvent(self, event):
        '''reimplemented to provide drag events'''
        print 'In DefaultLabelWidget.mousePressEvent'
        Qt.QLabel.mousePressEvent(self, event)
        if event.button() == Qt.Qt.LeftButton:
            self.dragStartPosition = event.pos()
            
    def mouseMoveEvent(self, event):
        '''reimplemented to provide drag events'''
        if not event.buttons() & Qt.Qt.LeftButton:
            return
        mimeData = Qt.QMimeData()
        txt = str(self._drageventcallback())
        print 'Draggable.mouseMoveEvent',self._drageventcallback,txt
        mimeData.setText(txt)
        drag = Qt.QDrag(self)
        drag.setMimeData(mimeData)
        drag.setHotSpot(event.pos() - self.rect().topLeft())
        dropAction = drag.start(Qt.Qt.CopyAction)
    
class DraggableLabel(Draggable,Qt.QLabel):
    def __init__(self,parent=None,text=''):
        Qt.QLabel.__init__(self,text)
        Draggable.__init__(self)
        
def YesNoDialog(title,message):
    v = Qt.QMessageBox.warning(None,title,message, \
    Qt.QMessageBox.Yes|Qt.QMessageBox.No);
    return v == Qt.QMessageBox.Yes


class MyTaurusComponent(TaurusBaseComponent,Qt.QObject):
    def __init__(self,name='test'):
        Qt.QObject.__init__(self)
        TaurusBaseComponent.__init__(self,name)
        self.my_handler = f.printf
        
    def fireEvent(self, evt_src=None, evt_type=None, evt_value=None):
        if self._eventBufferPeriod:
            # If we have an active event buffer delay, store the event...
            with self._eventsBufferLock:
                self._bufferedEvents[(evt_src, evt_type)] =  (evt_src, evt_type, 
                                                              evt_value) 
        else:
            # if we are not buffering, directly emit the signal
            try: 
                self.getSignaller().emit(Qt.SIGNAL('taurusEvent'), evt_src, evt_type, evt_value)
            except: 
                self.error('%s.fireEvent(...) failed!'%type(self))
                self.error(f.getLastException())

    def setMyHandler(self,handler):
        self.my_handler = handler                
    @f.Catched
    def handleEvent(self,*args):
        #self.error('in handleEvent(%s,%s)'%(str(args[0]),len(args)))
        self.my_handler(*args)
        
class addCustomPanel2Gui(object):
    """
    this Class is used to add new panel to existing TaurusGUI with
    the same Context.
    Each extra_panel class should have a 'getPaneldescription' and
    getDefaultIcon methods.
    
    Pass a dictionary like this:
       EXTRA_PANELS['VaccaProperties'] = {
        'class': vacca.VaccaPropTable,
        'icon': ':/places/network-server.svg',
        }
    """

    def __init__(self, extra_panels= None):
        print '>>'*80
        print '>>'*80
        print '>>'*80

        exist_instance = False
        try:
            #find actual Taurusgui Instance
            
            app = Qt.QApplication.instance()
            assert app
        except:
            traceback.print_exc()
            return
        
        try:
            taurusgui = None
            widgets = app.allWidgets()
            for widget in widgets:
                if isinstance(widget, TaurusGui):
                    taurusgui = widget
            #taurusgui = get_taurus_gui() #get_main_window() does not work here!?

            #Launch method
            def _launchPanel(panelName, panelClass, *args):
                #print args
                #print panelName, panelClass
                #print "Launch Panel"

                exist = False

                #find if exists instance with this Widget
                for widget in app.allWidgets():
                    widgetType = str(type(widget))
                    if str(panelClass) in widgetType:

                        #If exist set Flag to know it
                        exist = True
                        taurusgui.setFocusToPanel(panelName)
                        break

                if not exist:
                    try:
                        taurusgui.createCustomPanel(panelClass.getPanelDescription(panelName))
                    except:
                        taurusgui.createCustomPanel(PanelDescription(panelName,panelClass.__module__+'.'+panelClass.__name__))

            def launchPanel(panelName, panelClass):
                return lambda args: _launchPanel(panelName, panelClass, args)

            for panel in extra_panels.keys():
              action = None
              try:
                d = extra_panels[panel]
                class_Panel = d.get('class',d.get('classname'))
                print panel,class_Panel,d
                try:
                    icon = extra_panels[panel].get('icon',None) or class_Panel.getDefaultIcon()
                except Exception,e:
                    print('Icon exception in addCustomPanel2Gui(%s)!! %s'%(panel,e))
                    icon = ':/places/network-server.svg'
                final_icon_url = icon if fandango.matchCl('^[\:/]',icon) else wdir('vacca/'+icon)

                #button = TaurusLauncherButton(
                #    widget=class_Panel.getPanelDescription(panel))
                #taurusgui.jorgsBar.addWidget(button)
                FUNCTION_TYPE = type(launchPanel)
                METHOD_TYPE = type(MyTaurusComponent.fireEvent)
                
                if 'VaccaAction' in str(class_Panel):
                    import vacca.panel
                    action = vacca.panel.VaccaAction(parent=taurusgui)
                else:
                    action = Qt.QAction(Qt.QIcon(final_icon_url),
                        panel, taurusgui)
                    if d.get('name'): action.setText(d['name'])
                    
                if isinstance(action,Qt.QAction):
                    #func = launchPanel(panel, class_Panel)
                    if isinstance(class_Panel,(FUNCTION_TYPE,METHOD_TYPE)):
                        func = class_Panel
                    else:
                        func = lambda args,panelName=panel,panelClass=class_Panel: \
                            _launchPanel(panelName,panelClass, args)
                    Qt.QObject.connect(action, Qt.SIGNAL("triggered(bool)"), func)
                    taurusgui.jorgsBar.addAction(action)
                else:
                    if d.get('model'): action.setModel(d['model'])
                    taurusgui.jorgsBar.addWidget(action)

                print "%s Added in jorgsBar"%(panel)
              except:
                print('Error creating %s: %s,%s'%(panel,extra_panels[panel],action))
                traceback.print_exc()
        except:
            traceback.print_exc()

from .doc import get_autodoc
__doc__ = get_autodoc(__name__,vars())
